"""Module for generating some sample datasets.

"""

import math
import string

import networkx as nx
import numpy as np
import pandas as pd
import scipy.stats as ss
from numpy.random import choice, random
from sklearn.neural_network import MLPRegressor

from dowhy.utils.graph_operations import (
    add_edge,
    convert_to_undirected_graph,
    del_edge,
    find_predecessor,
    get_random_node_pair,
    get_simple_ordered_tree,
    is_connected,
)


def sigmoid(x):
    return 1 / (1 + math.exp(-x))


def convert_to_binary(x, stochastic=True):
    p = sigmoid(x)
    if stochastic:
        return choice([0, 1], 1, p=[1 - p, p])
    else:
        return int(p > 0.5)


def stochastically_convert_to_three_level_categorical(x):
    p = sigmoid(x)
    return choice([0, 1, 2], 1, p=[0.8 * (1 - p), 0.8 * p, 0.2])


def convert_to_categorical(arr, num_vars, num_discrete_vars, quantiles=[0.25, 0.5, 0.75], one_hot_encode=False):
    arr_with_dummy = arr.copy()
    # Below loop assumes that the last indices of W are alwawys converted to discrete
    for arr_index in range(num_vars - num_discrete_vars, num_vars):
        # one-hot encode discrete W
        arr_bins = np.quantile(arr[:, arr_index], q=quantiles)
        arr_categorical = np.digitize(arr[:, arr_index], bins=arr_bins)
        if one_hot_encode:
            dummy_vecs = np.eye(len(quantiles) + 1)[arr_categorical]
            arr_with_dummy = np.concatenate((arr_with_dummy, dummy_vecs), axis=1)
        else:
            arr_with_dummy = np.concatenate((arr_with_dummy, arr_categorical[:, np.newaxis]), axis=1)
    # Now deleting the old continuous value
    for arr_index in range(num_vars - 1, num_vars - num_discrete_vars - 1, -1):
        arr_with_dummy = np.delete(arr_with_dummy, arr_index, axis=1)
    return arr_with_dummy


def construct_col_names(name, num_vars, num_discrete_vars, num_discrete_levels, one_hot_encode):
    colnames = [(name + str(i)) for i in range(0, num_vars - num_discrete_vars)]
    if one_hot_encode:
        discrete_colnames = [
            name + str(i) + "_" + str(j)
            for i in range(num_vars - num_discrete_vars, num_vars)
            for j in range(0, num_discrete_levels)
        ]
        colnames = colnames + discrete_colnames
    else:
        colnames = colnames + [(name + str(i)) for i in range(num_vars - num_discrete_vars, num_vars)]

    return colnames


def linear_dataset(
    beta,
    num_common_causes,
    num_samples,
    num_instruments=0,
    num_effect_modifiers=0,
    num_treatments=1,
    num_frontdoor_variables=0,
    treatment_is_binary=True,
    treatment_is_category=False,
    outcome_is_binary=False,
    stochastic_discretization=True,
    num_discrete_common_causes=0,
    num_discrete_instruments=0,
    num_discrete_effect_modifiers=0,
    stddev_treatment_noise=1,
    stddev_outcome_noise=0.01,
    one_hot_encode=False,
):
    """
    Generate a synthetic dataset with a known effect size.

    This function generates a pandas dataFrame with num_samples records. The variables follow a naming convention where the first letter indicates its role in the causality graph and then a sequence number.

    :param beta: coefficient of the treatment(s) ('v?') in the generating equation of the outcome ('y').
    :type beta: int or list/ndarray of length num_treatments of type int
    :param num_common_causes: Number of variables affecting both the treatment and the outcome [w -> v; w -> y]
    :type num_common_causes: int
    :param num_samples: Number of records to generate
    :type num_samples: int
    :param num_instruments: Number of instrumental variables  [z -> v], defaults to 0
    :type num_instruments: int
    :param num_effect_modifiers: Number of effect modifiers, variables affecting only the outcome [x -> y], defaults to 0
    :type num_effect_modifiers: int
    :param num_treatments: Number of treatment variables [v], defaults to 1
    :type num_treatments : int
    :param num_frontdoor_variables : Number of frontdoor mediating variables [v -> FD -> y], defaults to  0
    :type num_frontdoor_variables: int
    :param treatment_is_binary: Cannot be True if treatment_is_category is True, defaults to True
    :type treatment_is_binary: bool
    :param treatment_is_category: Cannot be True if treatment_is_binary is True, defaults to False
    :type treatment_is_category: bool
    :param outcome_is_binary: defaults to False,
    :type outcome_is_binary: bool
    :param stochastic_discretization: if False, quartiles are used when discretised variables are specified. They can be hot encoded, defaults True
    :type stochastic_discretization: bool
    :param num_discrete_common_causes: Number of discrete common causes of the total num_common_causes, defaults to 0
    :type num_discrete_common_causes: int
    :param num_discrete_instruments: Number of discrete instrumental variables of the total num_instruments, defaults to 0
    :type num_discrete_instruments  : int
    :param num_discrete_effect_modifiers : Number of discrete effect modifiers of the total effect_modifiers, defaults to 0
    :type num_discrete_effect_modifiers: int
    :param stddev_treatment_noise : defaults to 1
    :type stddev_treatment_noise : float
    :param stddev_outcome_noise: defaults to 0.01
    :type stddev_outcome_noise: float
    :param one_hot_encode: defaults to False
    :type one_hot_encode: bool

    :returns: Dictionary with pandas dataFrame and few other metadata variables.
                        "df": pd.dataFrame
                        with num_samples records. The variables follow a naming convention were the first letter indicates its role in the causality graph and then a sequence number.

                    v variables - are the treatments. They can be binary or continuous. In the case of continuous abs(*beta*) defines thier magnitude;

                    y - is the outcome variable. The generating equation is,
                     y = normal(0, stddev_outcome_noise) + t @ beta [where @ is a numpy matrix multiplication allowing for beta be a vector]

                    W variables - commonly cause both the treatment and the outcome and are iid. if continuous, they are Norm(mu = Unif(-1,1), sigma = 1)

                    Z variables - Instrument variables. Each one affects all treatments. i.e. if there is one instrument and two treatments then z0->v0, z0->v1

                    X variables - effect modifiers. If continuous, they are Norm(mu = Unif(-1,1), sigma = 1)

                    FD variables - Front door variables, v0->FD0->y

            "treatment_name": str/list(str)
            "outcome_name": str
            "common_causes_names": str/list(str)
            "instrument_names": str/list(str)
            "effect_modifier_names": str/list(str)
            "frontdoor_variables_names": str/list(str)
            "dot_graph": dot_graph,
            "gml_graph": gml_graph,
            "ate": float, the true ate in the dataset
    :rtype: dict

    Examples
    ********
    .. code-block:: python
            import networkx as nx
            import matplotlib.pyplot as plt
            import pandas as pd
            import numpy as np
            import dowhy.datasets

            def plot_gml(gml_graph):
                    G = nx.parse_gml(gml_graph)
                    pos=nx.spring_layout(G)
                    nx.draw_networkx(G, pos, with_labels=True, node_size=1000, node_color="darkorange")
                    return(plt.show())

            def describe_synthetic_data(synthetic_data):
                    if (synthetic_data['gml_graph'] != None) :
                    plot_gml(synthetic_data["gml_graph"])
                    synthetic_data_df=synthetic_data["df"]
                    print('------- Variables --------')
                    print('Treatment vars:'      , synthetic_data['treatment_name'])
                    print('Outcome vars:'        , synthetic_data['outcome_name'])
                    print('Common causes vars:'  , synthetic_data['common_causes_names'])
                    print('Instrument vars:'     , synthetic_data['instrument_names'])
                    print('Effect Modifier vars:', synthetic_data['effect_modifier_names'])
                    print('Frontdoor vars:'      , synthetic_data['frontdoor_variables_names'])
                    print('Treatment vars:', synthetic_data['outcome_name'])
                    print('-------- Corr -------')
                    print(synthetic_data_df.corr())
                    print('------- Head --------')
                    return(synthetic_data_df)

            # create a dataset with 10 observations one binary treatment and a continuous outcome affected by one common cause
            synthetic_data = dowhy.datasets.linear_dataset(beta = 100,
                    num_common_causes = 1,
                    num_samples =10
                    )
            describe_synthetic_data(synthetic_data).head()

            # Two continuous treatments, no common cause, an instrumental variable and two effect modifiers - linearly added appropriately
            synthetic_data = dowhy.datasets.linear_dataset(
                    beta                          = 100,
                    num_common_causes             = 0,
                    num_samples                   = 20,
                    num_instruments               = 1,
                    num_effect_modifiers          = 2,
                    num_treatments                = 2,
                    num_frontdoor_variables       = 0,
                    treatment_is_binary           = False,
                    treatment_is_category         = False,
                    outcome_is_binary             = False,
                    stochastic_discretization     = True,
                    num_discrete_common_causes    = 0,
                    num_discrete_instruments      = 0,
                    num_discrete_effect_modifiers = 0,
                    stddev_treatment_noise        = 1,
                    stddev_outcome_noise          = 0.01,
                    one_hot_encode                = False
                    )
            describe_synthetic_data(synthetic_data).head()

            # One Hot Encoding
            synthetic_data = dowhy.datasets.linear_dataset(
                    beta                          = 100,
                    num_common_causes             =   2,
                    num_samples                   =  20,
                    num_instruments               =   1,
                    num_effect_modifiers          =   1,
                    num_treatments                =   1,
                    num_frontdoor_variables       =   1,
                    treatment_is_binary           = False,
                    treatment_is_category         = False,
                    outcome_is_binary             = False,
                    stochastic_discretization     = True,
                    num_discrete_common_causes    = 1, #of the total num_common_causes
                    num_discrete_instruments      = 1,
                    num_discrete_effect_modifiers = 1,
                    stddev_treatment_noise        = 1,
                    stddev_outcome_noise          = 0.01,
                    one_hot_encode                = True
                    )
                    describe_synthetic_data(synthetic_data).head()
    """
    assert not (treatment_is_binary and treatment_is_category)
    W, X, Z, FD, c1, c2, ce, cz, cfd1, cfd2 = [None] * 10
    W_with_dummy, X_with_categorical = (None, None)
    beta = float(beta)
    # Making beta an array
    if type(beta) not in [list, np.ndarray]:
        beta = np.repeat(beta, num_treatments)
    num_cont_common_causes = num_common_causes - num_discrete_common_causes
    num_cont_instruments = num_instruments - num_discrete_instruments
    num_cont_effect_modifiers = num_effect_modifiers - num_discrete_effect_modifiers
    if num_common_causes > 0:
        range_c1 = 0.5 + max(abs(beta)) * 0.5
        range_c2 = 0.5 + max(abs(beta)) * 0.5
        means = np.random.uniform(-1, 1, num_common_causes)
        cov_mat = np.diag(np.ones(num_common_causes))
        W = np.random.multivariate_normal(means, cov_mat, num_samples)
        W_with_dummy = convert_to_categorical(
            W, num_common_causes, num_discrete_common_causes, quantiles=[0.25, 0.5, 0.75], one_hot_encode=one_hot_encode
        )
        c1 = np.random.uniform(0, range_c1, (W_with_dummy.shape[1], num_treatments))
        c2 = np.random.uniform(0, range_c2, W_with_dummy.shape[1])

    if num_instruments > 0:
        range_cz = 1 + max(abs(beta))
        p = np.random.uniform(0, 1, num_instruments)
        Z = np.zeros((num_samples, num_instruments))
        for i in range(num_instruments):
            if (i % 2) == 0:
                Z[:, i] = np.random.binomial(n=1, p=p[i], size=num_samples)
            else:
                Z[:, i] = np.random.uniform(0, 1, size=num_samples)
        # TODO Ensure that we do not generate weak instruments
        cz = np.random.uniform(
            range_cz - (range_cz * 0.05), range_cz + (range_cz * 0.05), (num_instruments, num_treatments)
        )
    if num_effect_modifiers > 0:
        range_ce = 0.5 + max(abs(beta)) * 0.5
        means = np.random.uniform(-1, 1, num_effect_modifiers)
        cov_mat = np.diag(np.ones(num_effect_modifiers))
        X = np.random.multivariate_normal(means, cov_mat, num_samples)
        X_with_categorical = convert_to_categorical(
            X,
            num_effect_modifiers,
            num_discrete_effect_modifiers,
            quantiles=[0.25, 0.5, 0.75],
            one_hot_encode=one_hot_encode,
        )
        ce = np.random.uniform(0, range_ce, X_with_categorical.shape[1])
    # TODO - test all our methods with random noise added to covariates (instead of the stochastic treatment assignment)

    t = np.random.normal(0, stddev_treatment_noise, (num_samples, num_treatments))
    if num_common_causes > 0:
        t += W_with_dummy @ c1  # + np.random.normal(0, 0.01)
    if num_instruments > 0:
        t += Z @ cz
    # Converting treatment to binary if required
    if treatment_is_binary:
        t = np.vectorize(convert_to_binary)(t)
    elif treatment_is_category:
        t = np.vectorize(stochastically_convert_to_three_level_categorical)(t)

    # Generating frontdoor variables if asked for
    if num_frontdoor_variables > 0:
        range_cfd1 = max(abs(beta)) * 0.5
        range_cfd2 = max(abs(beta)) * 0.5
        cfd1 = np.random.uniform(0, range_cfd1, (num_treatments, num_frontdoor_variables))
        cfd2 = np.random.uniform(0, range_cfd2, num_frontdoor_variables)
        FD_noise = np.random.normal(0, 1, (num_samples, num_frontdoor_variables))
        FD = FD_noise
        FD += t @ cfd1
        if num_common_causes > 0:
            range_c1_frontdoor = range_c1 / 10.0
            c1_frontdoor = np.random.uniform(0, range_c1_frontdoor, (W_with_dummy.shape[1], num_frontdoor_variables))
            FD += W_with_dummy @ c1_frontdoor

    def _compute_y(t, W, X, FD, beta, c2, ce, cfd2, stddev_outcome_noise):
        y = np.random.normal(0, stddev_outcome_noise, num_samples)
        if num_frontdoor_variables > 0:
            y += FD @ cfd2
        else:
            # NOTE: We are assuming a linear relationship *even when t is categorical* and integer coded.
            # For categorical t, this example dataset has the effect size for category 2 being exactly
            # double the effect for category 1
            # This could be changed at this stage by one-hot encoding t and using a custom beta that
            # sets a different effect for each category {0, 1, 2}
            y += t @ beta
        if num_common_causes > 0:
            y += W @ c2
        if num_effect_modifiers > 0:
            y += (X @ ce) * np.prod(t, axis=1)
        if outcome_is_binary:
            y = np.vectorize(convert_to_binary)(y, stochastic_discretization)
        return y

    y = _compute_y(t, W_with_dummy, X_with_categorical, FD, beta, c2, ce, cfd2, stddev_outcome_noise)

    data = np.column_stack((t, y))
    if num_common_causes > 0:
        data = np.column_stack((W_with_dummy, data))
    if num_instruments > 0:
        data = np.column_stack((Z, data))
    if num_effect_modifiers > 0:
        data = np.column_stack((X_with_categorical, data))
    if num_frontdoor_variables > 0:
        data = np.column_stack((FD, data))

    # Computing ATE
    FD_T1, FD_T0 = None, None
    T1 = np.ones((num_samples, num_treatments))
    T0 = np.zeros((num_samples, num_treatments))
    if num_frontdoor_variables > 0:
        FD_T1 = FD_noise + (T1 @ cfd1)
        FD_T0 = FD_noise + (T0 @ cfd1)
    ate = np.mean(
        _compute_y(T1, W_with_dummy, X_with_categorical, FD_T1, beta, c2, ce, cfd2, stddev_outcome_noise)
        - _compute_y(T0, W_with_dummy, X_with_categorical, FD_T0, beta, c2, ce, cfd2, stddev_outcome_noise)
    )

    treatments = [("v" + str(i)) for i in range(0, num_treatments)]
    outcome = "y"
    # constructing column names for one-hot encoded discrete features
    common_causes = construct_col_names(
        "W", num_common_causes, num_discrete_common_causes, num_discrete_levels=4, one_hot_encode=one_hot_encode
    )
    instruments = [("Z" + str(i)) for i in range(0, num_instruments)]
    frontdoor_variables = [("FD" + str(i)) for i in range(0, num_frontdoor_variables)]
    effect_modifiers = construct_col_names(
        "X", num_effect_modifiers, num_discrete_effect_modifiers, num_discrete_levels=4, one_hot_encode=one_hot_encode
    )
    other_variables = None
    col_names = frontdoor_variables + effect_modifiers + instruments + common_causes + treatments + [outcome]
    data = pd.DataFrame(data, columns=col_names)
    # Specifying the correct dtypes
    if treatment_is_binary:
        data = data.astype({tname: "bool" for tname in treatments}, copy=False)
    elif treatment_is_category:
        data = data.astype({tname: "category" for tname in treatments}, copy=False)
    if outcome_is_binary:
        data = data.astype({outcome: "bool"}, copy=False)
    if num_discrete_common_causes > 0 and not one_hot_encode:
        data = data.astype({wname: "int64" for wname in common_causes[num_cont_common_causes:]}, copy=False)
        data = data.astype({wname: "category" for wname in common_causes[num_cont_common_causes:]}, copy=False)
    if num_discrete_effect_modifiers > 0 and not one_hot_encode:
        data = data.astype({emodname: "int64" for emodname in effect_modifiers[num_cont_effect_modifiers:]}, copy=False)
        data = data.astype(
            {emodname: "category" for emodname in effect_modifiers[num_cont_effect_modifiers:]}, copy=False
        )

    # Now specifying the corresponding graph strings
    dot_graph = create_dot_graph(treatments, outcome, common_causes, instruments, effect_modifiers, frontdoor_variables)
    # Now writing the gml graph
    gml_graph = create_gml_graph(treatments, outcome, common_causes, instruments, effect_modifiers, frontdoor_variables)
    ret_dict = {
        "df": data,
        "treatment_name": treatments,
        "outcome_name": outcome,
        "common_causes_names": common_causes,
        "instrument_names": instruments,
        "effect_modifier_names": effect_modifiers,
        "frontdoor_variables_names": frontdoor_variables,
        "dot_graph": dot_graph,
        "gml_graph": gml_graph,
        "ate": ate,
    }
    return ret_dict


def simple_iv_dataset(beta, num_samples, num_treatments=1, treatment_is_binary=True, outcome_is_binary=False):
    """Simple instrumental variable dataset with a single IV and a single confounder."""
    W, Z, c1, c2, cz = [None] * 5
    num_instruments = 1
    num_common_causes = 1
    beta = float(beta)
    # Making beta an array
    if type(beta) not in [list, np.ndarray]:
        beta = np.repeat(beta, num_treatments)

    c1 = np.random.uniform(0, 1, (num_common_causes, num_treatments))
    c2 = np.random.uniform(0, 1, num_common_causes)
    range_cz = 1 + max(abs(beta))  # cz is much higher than c1 and c2
    cz = np.random.uniform(
        range_cz - (range_cz * 0.05), range_cz + (range_cz * 0.05), (num_instruments, num_treatments)
    )
    W = np.random.uniform(0, 1, (num_samples, num_common_causes))
    Z = np.random.normal(0, 1, (num_samples, num_instruments))
    t = np.random.normal(0, 1, (num_samples, num_treatments)) + Z @ cz + W @ c1
    if treatment_is_binary:
        t = np.vectorize(convert_to_binary)(t)

    def _compute_y(t, W, beta, c2):
        y = t @ beta + W @ c2
        return y

    y = _compute_y(t, W, beta, c2)

    # creating data frame
    data = np.column_stack((Z, W, t, y))
    treatments = [("v" + str(i)) for i in range(0, num_treatments)]
    outcome = "y"
    common_causes = [("W" + str(i)) for i in range(0, num_common_causes)]
    ate = np.mean(
        _compute_y(np.ones((num_samples, num_treatments)), W, beta, c2)
        - _compute_y(np.zeros((num_samples, num_treatments)), W, beta, c2)
    )
    instruments = [("Z" + str(i)) for i in range(0, num_instruments)]
    other_variables = None
    col_names = instruments + common_causes + treatments + [outcome]
    data = pd.DataFrame(data, columns=col_names)

    # Specifying the correct dtypes
    if treatment_is_binary:
        data = data.astype({tname: "bool" for tname in treatments}, copy=False)
    if outcome_is_binary:
        data = data.astype({outcome: "bool"}, copy=False)

    # Now specifying the corresponding graph strings
    dot_graph = create_dot_graph(treatments, outcome, common_causes, instruments)
    # Now writing the gml graph
    gml_graph = create_gml_graph(treatments, outcome, common_causes, instruments)
    ret_dict = {
        "df": data,
        "treatment_name": treatments,
        "outcome_name": outcome,
        "common_causes_names": common_causes,
        "instrument_names": instruments,
        "effect_modifier_names": None,
        "dot_graph": dot_graph,
        "gml_graph": gml_graph,
        "ate": ate,
    }
    return ret_dict


def create_dot_graph(treatments, outcome, common_causes, instruments, effect_modifiers=[], frontdoor_variables=[]):
    dot_graph = "digraph {"
    for currt in treatments:
        if len(frontdoor_variables) == 0:
            dot_graph += "{0}->{1};".format(currt, outcome)
        dot_graph += " ".join([v + "-> " + currt + ";" for v in common_causes])
        dot_graph += " ".join([v + "-> " + currt + ";" for v in instruments])
        dot_graph += " ".join([currt + "-> " + v + ";" for v in frontdoor_variables])

    dot_graph += " ".join([v + "-> " + outcome + ";" for v in common_causes])
    dot_graph += " ".join([v + "-> " + outcome + ";" for v in effect_modifiers])
    dot_graph += " ".join([v + "-> " + outcome + ";" for v in frontdoor_variables])
    dot_graph = dot_graph + "}"
    # Adding edges between common causes and the frontdoor mediator
    for v1 in common_causes:
        dot_graph += " ".join([v1 + "-> " + v2 + ";" for v2 in frontdoor_variables])
    return dot_graph


def create_gml_graph(treatments, outcome, common_causes, instruments, effect_modifiers=[], frontdoor_variables=[]):
    gml_graph = ("graph[directed 1" 'node[ id "{0}" label "{0}"]').format(outcome)

    gml_graph += " ".join(['node[ id "{0}" label "{0}"]'.format(v) for v in common_causes])
    gml_graph += " ".join(['node[ id "{0}" label "{0}"]'.format(v) for v in instruments])
    gml_graph += " ".join(['node[ id "{0}" label "{0}"]'.format(v) for v in frontdoor_variables])
    for currt in treatments:
        gml_graph += ('node[ id "{0}" label "{0}"]').format(currt)
        if len(frontdoor_variables) == 0:
            gml_graph += 'edge[source "{0}" target "{1}"]'.format(currt, outcome)
        gml_graph += " ".join(['edge[ source "{0}" target "{1}"]'.format(v, currt) for v in common_causes])
        gml_graph += " ".join(['edge[ source "{0}" target "{1}"]'.format(v, currt) for v in instruments])
        gml_graph += " ".join(['edge[ source "{0}" target "{1}"]'.format(currt, v) for v in frontdoor_variables])

    gml_graph = gml_graph + " ".join(['edge[ source "{0}" target "{1}"]'.format(v, outcome) for v in common_causes])
    gml_graph = gml_graph + " ".join(
        ['node[ id "{0}" label "{0}"] edge[ source "{0}" target "{1}"]'.format(v, outcome) for v in effect_modifiers]
    )
    gml_graph = gml_graph + " ".join(
        ['edge[ source "{0}" target "{1}"]'.format(v, outcome) for v in frontdoor_variables]
    )
    for v1 in common_causes:
        gml_graph = gml_graph + " ".join(
            ['edge[ source "{0}" target "{1}"]'.format(v1, v2) for v2 in frontdoor_variables]
        )
    gml_graph = gml_graph + "]"
    return gml_graph


def xy_dataset(num_samples, effect=True, num_common_causes=1, is_linear=True, sd_error=1):
    treatment = "Treatment"
    outcome = "Outcome"
    common_causes = ["w" + str(i) for i in range(num_common_causes)]
    time_var = "s"
    # Error terms
    E1 = np.random.normal(loc=0, scale=sd_error, size=num_samples)
    E2 = np.random.normal(loc=0, scale=sd_error, size=num_samples)

    S = np.random.uniform(0, 10, num_samples)
    T1 = 4 - (S - 3) * (S - 3)
    T1[S >= 5] = 0
    T2 = (S - 7) * (S - 7) - 4
    T2[S <= 5] = 0
    W0 = T1 + T2  # hidden confounder
    tterm, yterm = 0, 0
    if num_common_causes > 1:
        means = np.random.uniform(-1, 1, num_common_causes - 1)
        cov_mat = np.diag(np.ones(num_common_causes - 1))
        otherW = np.random.multivariate_normal(means, cov_mat, num_samples)
        c1 = np.random.uniform(0, 1, (otherW.shape[1], 1))
        c2 = np.random.uniform(0, 1, (otherW.shape[1], 1))
        tterm = (otherW @ c1)[:, 0]
        yterm = (otherW @ c2)[:, 0]

    if is_linear:
        V = 6 + W0 + tterm + E1
        Y = 6 + W0 + yterm + E2  # + (V-8)*(V-8)
        if effect:
            Y += V
        else:
            Y += 6 + W0
    else:
        V = 6 + W0 * W0 + tterm + E1
        Y = 6 + W0 * W0 + yterm + E2  # + (V-8)*(V-8)
        if effect:
            Y += V  # /20 # divide by 10 to scale the value of Y to be comparable to V
        else:
            Y += 6 + W0
    # else:
    #    V = 6 + W0 + tterm + E1
    #    Y = 12 + W0*W0 + W0*W0 + yterm + E2  # E2_new
    dat = {treatment: V, outcome: Y, common_causes[0]: W0, time_var: S}
    if num_common_causes > 1:
        for i in range(otherW.shape[1]):
            dat[common_causes[i + 1]] = otherW[:, i]
    data = pd.DataFrame(data=dat)
    ret_dict = {
        "df": data,
        "treatment_name": treatment,
        "outcome_name": outcome,
        "common_causes_names": common_causes,
        "time_val": time_var,
        "instrument_names": None,
        "dot_graph": None,
        "gml_graph": None,
        "ate": None,
    }
    return ret_dict


def create_discrete_column(num_samples, std_dev=1):
    # Generating a random normal distribution of integers
    x = np.arange(-5, 6)
    xU, xL = x + 0.5, x - 0.5
    prob = ss.norm.cdf(xU, scale=std_dev) - ss.norm.cdf(
        xL, scale=std_dev
    )  # probability of selecting a number x is p(x-0.5 < x < x+0.5) where x is a normal random variable with mean 0 and standard deviation std_dev
    prob = prob / prob.sum()  # normalize the probabilities so their sum is 1
    nums = choice(a=x, size=num_samples, p=prob)  # pick up an element
    return nums


def convert_continuous_to_discrete(arr):
    return arr.astype(int)


def generate_random_graph(n, max_iter=10):
    """
    Function to generate random Directed Acyclic Graph
    :param n: number of nodes in the graph
    :param max_iter: number of iterations to create graph

    :returns: Directed Acyclic Graph
    See: https://datascience.oneoffcoder.com/generate-random-bbn.html
    """
    g = get_simple_ordered_tree(n)
    for it in range(max_iter):
        i, j = get_random_node_pair(n)
        if g.has_edge(i, j) is True:
            del_edge(i, j, g)
        else:
            add_edge(i, j, g)
    return g


def dataset_from_random_graph(
    num_vars, num_samples=1000, prob_edge=0.3, random_seed=100, prob_type_of_data=(0.333, 0.333, 0.334)
):
    """
    This function generates a dataset with discrete and continuous kinds of variables.
    It creates a random graph and models the variables linearly according to the relations in the graph.

    :param num_vars: Number of variables in the dataset
    :param num_samples: Number of samples in the dataset
    :param prob_edge : Probability of an edge between two random nodes in a graph
    :param random_seed: Seed for generating random graph
    :param prob_type_of_data : 3-element tuple containing the probability of data being discrete, binary and continuous respectively.
    :returns ret_dict : dictionary with information like dataframe, outcome, treatment, graph string and continuous, discrete and binary columns
    """
    assert sum(list(prob_type_of_data)) == 1.0
    np.random.seed(100)
    DAG = generate_random_graph(n=num_vars)
    mapping = dict(zip(DAG, string.ascii_lowercase))
    DAG = nx.relabel_nodes(DAG, mapping)
    all_nodes = list(DAG.nodes)
    all_nodes.sort()
    num_nodes = len(all_nodes)
    changed = dict()
    discrete_cols = []
    continuous_cols = []
    binary_cols = []
    random_numbers_array = np.random.rand(
        num_nodes
    )  # Random numbers between 0 to 1 to decide if that particular node will be discrete or continuous

    for node in all_nodes:
        changed[node] = False
    df = pd.DataFrame()
    currset = list()
    counter = 0

    # Generating data for nodes which have no incoming edges
    for node in all_nodes:
        if DAG.in_degree(node) == 0:
            x = random_numbers_array[counter]
            counter += 1
            if x <= prob_type_of_data[0]:
                df[node] = create_discrete_column(num_samples)  # Generating discrete data
                discrete_cols.append(node)
            elif x <= prob_type_of_data[0] + prob_type_of_data[1]:
                df[node] = np.random.normal(0, 1, num_samples)  # Generating continuous data
                continuous_cols.append(node)
            else:
                nums = np.random.normal(0, 1, num_samples)
                df[node] = np.vectorize(convert_to_binary)(nums)  # Generating binary data
                discrete_cols.append(node)
                binary_cols.append(node)
            successors = list(DAG.successors(node))  # Storing immediate successors for next level data generation
            successors.sort()
            currset.extend(successors)
            changed[node] = True

    # "currset" variable currently has all the successors of the nodes which had no incoming edges
    while len(currset) > 0:
        cs = list()  # Variable to store immediate children of nodes present in "currset"
        for node in currset:
            predecessors = list(
                DAG.predecessors(node)
            )  # Getting all the parent nodes on which current "node" depends on
            if changed[node] == False and all(
                changed[x] == True for x in predecessors
            ):  # Check if current "node" has not been processed yet and if all the parent nodes have been processed
                successors = list(DAG.successors(node))
                successors.sort()
                cs.extend(successors)  # Storing immediate children for next level data generation
                X = df[predecessors].to_numpy()  # Using parent nodes data
                c = np.random.uniform(0, 1, len(predecessors))
                t = np.random.normal(0, 1, num_samples) + X @ c  # Using Linear Regression to generate data
                changed[node] = True
                x = random_numbers_array[counter]
                counter += 1
                if x <= prob_type_of_data[0]:
                    df[node] = convert_continuous_to_discrete(t)
                    discrete_cols.append(node)
                elif x <= prob_type_of_data[0] + prob_type_of_data[1]:
                    df[node] = t
                    continuous_cols.append(node)
                else:
                    nums = np.random.normal(0, 1, num_samples)
                    df[node] = np.vectorize(convert_to_binary)(nums)
                    discrete_cols.append(node)
                    binary_cols.append(node)
        currset = cs

    outcome = None
    for node in all_nodes:
        if DAG.out_degree(node) == 0:
            outcome = node  # Node which has no successors is outcome
            break

    treatments = list()
    for node in all_nodes:
        if DAG.in_degree(node) > 0:
            children = list(DAG.successors(node))
            if outcome in children:
                treatments.append(node)  # Node which causes outcome is treatment

    gml_str = "\n".join(nx.generate_gml(DAG))
    ret_dict = {
        "df": df,
        "outcome_name": outcome,
        "treatment_name": treatments,
        "gml_graph": gml_str,
        "discrete_columns": discrete_cols,
        "continuous_columns": continuous_cols,
        "binary_columns": binary_cols,
    }
    return ret_dict


def partially_linear_dataset(
    beta,
    num_common_causes,
    num_unobserved_common_causes=0,
    strength_unobserved_confounding=1,
    num_samples=500,
    num_treatments=1,
    treatment_is_binary=True,
    treatment_is_category=False,
    outcome_is_binary=False,
    stochastic_discretization=True,
    num_discrete_common_causes=0,
    stddev_treatment_noise=1,
    stddev_outcome_noise=0,
    one_hot_encode=False,
    training_sample_size=10,
    random_state=0,
):
    assert not (treatment_is_binary and treatment_is_category)
    num_outcomes = 1
    beta = float(beta)
    # Making beta an array
    if type(beta) not in [list, np.ndarray]:
        beta = np.repeat(beta, num_treatments)

    num_cont_common_causes = num_common_causes - num_discrete_common_causes

    if num_common_causes > 0:
        range_c1 = 0.5 + max(abs(beta)) * 0.5
        means = np.random.uniform(-1, 1, num_common_causes)
        cov_mat = np.diag(np.ones(num_common_causes))
        W = np.random.multivariate_normal(means, cov_mat, num_samples)
        W_with_dummy = convert_to_categorical(
            W, num_common_causes, num_discrete_common_causes, quantiles=[0.25, 0.5, 0.75], one_hot_encode=one_hot_encode
        )
        c1 = np.random.uniform(0, range_c1, (W_with_dummy.shape[1], num_treatments))
        # assuming that all unobserved common causes are numerical and are not affected by one hot encoding
        if num_unobserved_common_causes > 0:
            c1[:num_unobserved_common_causes] = c1[:num_unobserved_common_causes] * strength_unobserved_confounding
            for i in range(num_unobserved_common_causes, W_with_dummy.shape[1]):
                c1[i] = c1[i] * strength_unobserved_confounding / (2**i)

    # Creating a NN to simulate the nuisance function
    hidden_layer_arch = (50, 50, 50)
    neural_network = MLPRegressor(random_state=random_state, hidden_layer_sizes=hidden_layer_arch)
    x = np.random.randn(training_sample_size, W_with_dummy.shape[1])
    y = np.random.randn(training_sample_size, num_outcomes)
    neural_network.fit(x, np.ravel(y))

    T = np.random.normal(0, stddev_treatment_noise, (num_samples, num_treatments))
    T += W_with_dummy @ c1
    if treatment_is_binary:
        T = np.vectorize(convert_to_binary)(x=T)

    # strength of unobserved confounding
    strength_vec = np.ones(W_with_dummy.shape[1])
    strength_vec[:num_unobserved_common_causes] = (
        strength_vec[:num_unobserved_common_causes] * strength_unobserved_confounding
    )
    for i in range(num_unobserved_common_causes, W_with_dummy.shape[1]):
        strength_vec[i] = strength_vec[i] * strength_unobserved_confounding / (2**i)

    def _compute_y(T, W, beta, nn, stddev_outcome_noise):
        f_x = nn.predict(W)
        y = np.random.normal(0, stddev_outcome_noise, num_samples)
        y += T @ beta
        y += f_x
        if outcome_is_binary:
            y = np.vectorize(convert_to_binary)(y, stochastic_discretization)
        return y

    Y = _compute_y(T, W_with_dummy * strength_vec, beta, neural_network, stddev_outcome_noise)

    data = np.column_stack((T, Y))
    if num_common_causes > 0:
        data = np.column_stack((W_with_dummy, data))

    # Computing ATE
    T1 = np.ones((num_samples, num_treatments))
    T0 = np.zeros((num_samples, num_treatments))
    ate = np.mean(
        _compute_y(T1, W_with_dummy, beta, neural_network, stddev_outcome_noise)
        - _compute_y(T0, W_with_dummy, beta, neural_network, stddev_outcome_noise)
    )

    treatments = [("v" + str(i)) for i in range(0, num_treatments)]
    outcome = "y"
    common_causes = construct_col_names(
        "W", num_common_causes, num_discrete_common_causes, num_discrete_levels=4, one_hot_encode=one_hot_encode
    )
    col_names = common_causes + treatments + [outcome]
    data = pd.DataFrame(data, columns=col_names)
    # Specifying the correct dtypes
    if treatment_is_binary:
        data = data.astype({tname: "bool" for tname in treatments}, copy=False)
    elif treatment_is_category:
        data = data.astype({tname: "category" for tname in treatments}, copy=False)
    if outcome_is_binary:
        data = data.astype({outcome: "bool"}, copy=False)
    if num_discrete_common_causes > 0 and not one_hot_encode:
        data = data.astype({wname: "int64" for wname in common_causes[num_cont_common_causes:]}, copy=False)
        data = data.astype({wname: "category" for wname in common_causes[num_cont_common_causes:]}, copy=False)
    dot_graph = create_dot_graph(treatments, outcome, common_causes, instruments=[])
    # Now writing the gml graph
    gml_graph = create_gml_graph(treatments, outcome, common_causes, instruments=[])
    ret_dict = {
        "df": data,
        "treatment_name": treatments,
        "outcome_name": outcome,
        "common_causes_names": common_causes,
        "dot_graph": dot_graph,
        "gml_graph": gml_graph,
        "ate": ate,
    }
    return ret_dict


def lalonde_dataset() -> pd.DataFrame:
    """Downloads and returns the Lalonde dataset from https://users.nber.org/~rdehejia/nswdata2.html"""
    # The following code for loading the Lalonde dataset was copied from
    # https://github.com/wayfair/pylift/blob/5afc9088e96f25672423663f5c9b4bb889b4dfc0/examples/Lalonde/Lalonde_sample.ipynb?short_path=b1d451f#L94-L99).
    #
    # Copyright 2018, Wayfair, Inc.
    #
    # Redistribution and use in source and binary forms, with or without modification, are permitted provided that
    # the following conditions are met:
    #
    # 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the
    #    following disclaimer.
    #
    # 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
    #    following disclaimer in the documentation and/or other materials provided with the distribution.
    #
    # THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
    # WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
    # PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
    # DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    # PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
    # CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
    # OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
    # DAMAGE.
    cols = ["treat", "age", "educ", "black", "hisp", "married", "nodegr", "re74", "re75", "re78"]
    control = pd.read_csv(
        "https://www.nber.org/~rdehejia/data/nswre74_control.txt", sep="\\s+", header=None, names=cols
    )
    treated = pd.read_csv(
        "https://www.nber.org/~rdehejia/data/nswre74_treated.txt", sep="\\s+", header=None, names=cols
    )
    lalonde = pd.concat([control, treated], ignore_index=True).astype({"treat": "bool"}, copy=False)
    lalonde["u74"] = np.where(lalonde["re74"] == 0, 1.0, 0.0)
    lalonde["u75"] = np.where(lalonde["re75"] == 0, 1.0, 0.0)
    return lalonde



from typing import Any, List, Optional, Union

import numpy as np
import pandas as pd
from sklearn import linear_model
from sklearn.exceptions import NotFittedError

from dowhy.causal_estimator import CausalEstimator
from dowhy.causal_identifier import IdentifiedEstimand


class PropensityScoreEstimator(CausalEstimator):
    """
    Base class for estimators that estimate effects based on propensity of
    treatment assignment.

    Supports additional parameters as listed below.

    """

    def __init__(
        self,
        identified_estimand: IdentifiedEstimand,
        test_significance: bool = False,
        evaluate_effect_strength: bool = False,
        confidence_intervals: bool = False,
        num_null_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_STAT_TEST,
        num_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_CI,
        sample_size_fraction: int = CausalEstimator.DEFAULT_SAMPLE_SIZE_FRACTION,
        confidence_level: float = CausalEstimator.DEFAULT_CONFIDENCE_LEVEL,
        need_conditional_estimates: Union[bool, str] = "auto",
        num_quantiles_to_discretize_cont_cols: int = CausalEstimator.NUM_QUANTILES_TO_DISCRETIZE_CONT_COLS,
        propensity_score_model: Optional[Any] = None,
        propensity_score_column: str = "propensity_score",
        **kwargs,
    ):
        """
        :param identified_estimand: probability expression
            representing the target identified estimand to estimate.
        :param test_significance: Binary flag or a string indicating whether to test significance and by which method. All estimators support test_significance="bootstrap" that estimates a p-value for the obtained estimate using the bootstrap method. Individual estimators can override this to support custom testing methods. The bootstrap method supports an optional parameter, num_null_simulations. If False, no testing is done. If True, significance of the estimate is tested using the custom method if available, otherwise by bootstrap.
        :param evaluate_effect_strength: (Experimental) whether to evaluate the strength of effect
        :param confidence_intervals: Binary flag or a string indicating whether the confidence intervals should be computed and which method should be used. All methods support estimation of confidence intervals using the bootstrap method by using the parameter confidence_intervals="bootstrap". The bootstrap method takes in two arguments (num_simulations and sample_size_fraction) that can be optionally specified in the params dictionary. Estimators may also override this to implement their own confidence interval method. If this parameter is False, no confidence intervals are computed. If True, confidence intervals are computed by the estimator's specific method if available, otherwise through bootstrap
        :param num_null_simulations: The number of simulations for testing the
            statistical significance of the estimator
        :param num_simulations: The number of simulations for finding the
            confidence interval (and/or standard error) for a estimate
        :param sample_size_fraction: The size of the sample for the bootstrap
            estimator
        :param confidence_level: The confidence level of the confidence
            interval estimate
        :param need_conditional_estimates: Boolean flag indicating whether
            conditional estimates should be computed. Defaults to True if
            there are effect modifiers in the graph
        :param num_quantiles_to_discretize_cont_cols: The number of quantiles
            into which a numeric effect modifier is split, to enable
            estimation of conditional treatment effect over it.
        :param propensity_score_model: Model used to compute propensity score.
            Can be any classification model that supports fit() and
            predict_proba() methods. If None, LogisticRegression is used.
        :param propensity_score_column: Column name that stores the
            propensity score. Default='propensity_score'
        :param kwargs: (optional) Additional estimator-specific parameters
        """
        super().__init__(
            identified_estimand=identified_estimand,
            test_significance=test_significance,
            evaluate_effect_strength=evaluate_effect_strength,
            confidence_intervals=confidence_intervals,
            num_null_simulations=num_null_simulations,
            num_simulations=num_simulations,
            sample_size_fraction=sample_size_fraction,
            confidence_level=confidence_level,
            need_conditional_estimates=need_conditional_estimates,
            num_quantiles_to_discretize_cont_cols=num_quantiles_to_discretize_cont_cols,
            propensity_score_model=propensity_score_model,
            propensity_score_column=propensity_score_column,
            **kwargs,
        )

        # Enable the user to pass params for a custom propensity model
        self.propensity_score_model = propensity_score_model
        self.propensity_score_column = propensity_score_column

    def fit(
        self,
        data: pd.DataFrame,
        treatment_name: str,
        outcome_name: str,
        effect_modifier_names: Optional[List[str]] = None,
    ):
        """
        Fits the estimator with data for effect estimation
        :param data: data frame containing the data
        :param treatment: name of the treatment variable
        :param outcome: name of the outcome variable
        :param effect_modifiers: Variables on which to compute separate
                    effects, or return a heterogeneous effect function. Not all
                    methods support this currently.
        """
        self._set_data(data, treatment_name, outcome_name)
        self._set_effect_modifiers(effect_modifier_names)

        self.logger.debug("Back-door variables used:" + ",".join(self._target_estimand.get_backdoor_variables()))
        self._observed_common_causes_names = self._target_estimand.get_backdoor_variables()

        if self._observed_common_causes_names:
            self._observed_common_causes = self._data[self._observed_common_causes_names]
            # Convert the categorical variables into dummy/indicator variables
            # Basically, this gives a one hot encoding for each category
            # The first category is taken to be the base line.
            self._observed_common_causes = pd.get_dummies(self._observed_common_causes, drop_first=True)
        else:
            self._observed_common_causes = None
            error_msg = "No common causes/confounders present. Propensity score based methods are not applicable"
            self.logger.error(error_msg)
            raise Exception(error_msg)

        # Check if the treatment is one-dimensional
        if len(self._treatment_name) > 1:
            error_msg = str(self.__class__) + "cannot handle more than one treatment variable"
            raise Exception(error_msg)
        # Checking if the treatment is binary
        treatment_values = self._data[self._treatment_name[0]].astype(int).unique()
        if any([v not in [0, 1] for v in treatment_values]):
            error_msg = "Propensity score methods are applicable only for binary treatments"
            self.logger.error(error_msg)
            raise Exception(error_msg)

        if self.propensity_score_column not in self._data:
            if self.propensity_score_model is None:
                self.propensity_score_model = linear_model.LogisticRegression()
            treatment_reshaped = np.ravel(self._treatment)
            self.propensity_score_model.fit(self._observed_common_causes, treatment_reshaped)

        return self

    def estimate_propensity_score_column(self, data):
        try:
            data[self.propensity_score_column] = self.propensity_score_model.predict_proba(
                self._observed_common_causes
            )[:, 1]
        except NotFittedError:
            raise NotFittedError("Please fit the propensity score model before calling predict_proba")

    def construct_symbolic_estimator(self, estimand):
        """
        A symbolic string that conveys what each estimator does.
        For instance, linear regression is expressed as
        y ~ bx + e
        """
        raise NotImplementedError



import subprocess
import sys

import pytest
from pytest import mark

from dowhy import CausalModel
from dowhy.datasets import linear_dataset

causalml = pytest.importorskip("causalml")
from xgboost import XGBRegressor


@pytest.fixture
def init_data():
    data = linear_dataset(
        beta=10,
        num_common_causes=4,
        num_instruments=2,
        num_effect_modifiers=2,
        num_treatments=1,
        num_samples=1000,
        treatment_is_binary=True,
    )

    return data


@mark.use_fixtures("init_data")
class TestCausalmlEstimator:
    """
    To test the basic functionality of the CausalML estimators
    """

    def test_causalml_LRSRegressor(self, init_data):
        # Defined a linear dataset with a given set of properties
        data = init_data

        # Create a model that captures the same
        model = CausalModel(
            data=data["df"],
            treatment=data["treatment_name"],
            outcome=data["outcome_name"],
            effect_modifiers=data["effect_modifier_names"],
            graph=data["gml_graph"],
        )

        # Identify the effects within the model
        identified_estimand = model.identify_effect(proceed_when_unidentifiable=True)

        lr_estimate = model.estimate_effect(
            identified_estimand,
            method_name="backdoor.causalml.inference.meta.LRSRegressor",
            method_params={"init_params": {}},
        )

        print("The LR estimate obtained:")
        print(lr_estimate)

    def test_causalml_XGBTRegressor(self, init_data):
        # Defined a linear dataset with a given set of properties
        data = init_data

        # Create a model that captures the same
        model = CausalModel(
            data=data["df"],
            treatment=data["treatment_name"],
            outcome=data["outcome_name"],
            effect_modifiers=data["effect_modifier_names"],
            graph=data["gml_graph"],
        )

        # Identify the effects within the model
        identified_estimand = model.identify_effect(proceed_when_unidentifiable=True)

        xgbt_estimate = model.estimate_effect(
            identified_estimand,
            method_name="backdoor.causalml.inference.meta.XGBTRegressor",
            method_params={"init_params": {}},
        )

        print("The XGBT estimate obtained:")
        print(xgbt_estimate)

    def test_causalml_MLPTRegressor(self, init_data):
        # Defined a linear dataset with a given set of properties
        data = init_data

        # Create a model that captures the same
        model = CausalModel(
            data=data["df"],
            treatment=data["treatment_name"],
            outcome=data["outcome_name"],
            effect_modifiers=data["effect_modifier_names"],
            graph=data["gml_graph"],
        )

        # Identify the effects within the model
        identified_estimand = model.identify_effect(proceed_when_unidentifiable=True)

        mlpt_estimate = model.estimate_effect(
            identified_estimand,
            method_name="backdoor.causalml.inference.meta.MLPTRegressor",
            method_params={
                "init_params": {"hidden_layer_sizes": (10, 10), "learning_rate_init": 0.1, "early_stopping": True}
            },
        )

        print("The MLPT estimate obtained:")
        print(mlpt_estimate)

    def test_causalml_XLearner(self, init_data):
        # Defined a linear dataset with a given set of properties
        data = init_data

        # Create a model that captures the same
        model = CausalModel(
            data=data["df"],
            treatment=data["treatment_name"],
            outcome=data["outcome_name"],
            effect_modifiers=data["effect_modifier_names"],
            graph=data["gml_graph"],
        )

        # Identify the effects within the model
        identified_estimand = model.identify_effect(proceed_when_unidentifiable=True)

        xl_estimate = model.estimate_effect(
            identified_estimand,
            method_name="backdoor.causalml.inference.meta.BaseXRegressor",
            method_params={"init_params": {"learner": XGBRegressor()}},
        )

        print("The X Learner estimate obtained:")
        print(xl_estimate)

    def test_causalml_RLearner(self, init_data):
        # Defined a linear dataset with a given set of properties
        data = init_data

        # Create a model that captures the same
        model = CausalModel(
            data=data["df"],
            treatment=data["treatment_name"],
            outcome=data["outcome_name"],
            effect_modifiers=data["effect_modifier_names"],
            graph=data["gml_graph"],
        )

        # Identify the effects within the model
        identified_estimand = model.identify_effect(proceed_when_unidentifiable=True)

        rl_estimate = None

        try:
            rl_estimate = model.estimate_effect(
                identified_estimand,
                method_name="backdoor.causalml.inference.meta.BaseRRegressor",
                method_params={"init_params": {"learner": XGBRegressor()}},
            )
        except ValueError:
            print("Error with respect to the number of samples")

        print("The R Learner estimate obtained:")
        print(rl_estimate)



import numpy as np
from pytest import mark

from .base import TestRefuter


@mark.usefixtures("fixed_seed")
class TestDataSubsetRefuter(object):
    @mark.parametrize(["error_tolerance", "estimator_method"], [(0.01, "iv.instrumental_variable")])
    def test_refutation_data_subset_refuter_continuous(self, error_tolerance, estimator_method):
        refuter_tester = TestRefuter(error_tolerance, estimator_method, "data_subset_refuter")
        refuter_tester.continuous_treatment_testsuite()  # Run both

    @mark.parametrize(["error_tolerance", "estimator_method"], [(0.01, "backdoor.propensity_score_matching")])
    def test_refutation_data_subset_refuter_binary(self, error_tolerance, estimator_method):
        refuter_tester = TestRefuter(error_tolerance, estimator_method, "data_subset_refuter")
        refuter_tester.binary_treatment_testsuite(tests_to_run="atleast-one-common-cause")



import copy
from typing import List, Optional

import sympy as sp

from dowhy.utils.api import parse_state


class IdentifiedEstimand:

    """Class for storing a causal estimand, typically as a result of the identification step."""

    def __init__(
        self,
        identifier,  # This field will be deprecated in the future
        treatment_variable,
        outcome_variable,
        estimand_type=None,
        estimands=None,
        backdoor_variables=None,
        instrumental_variables=None,
        frontdoor_variables=None,
        mediator_variables=None,
        mediation_first_stage_confounders=None,
        mediation_second_stage_confounders=None,
        default_backdoor_id=None,
        identifier_method=None,
        no_directed_path=False,
    ):
        self.identifier = identifier
        self.treatment_variable = parse_state(treatment_variable)
        self.outcome_variable = parse_state(outcome_variable)
        self.backdoor_variables = backdoor_variables
        self.instrumental_variables = parse_state(instrumental_variables)
        self.frontdoor_variables = parse_state(frontdoor_variables)
        self.mediator_variables = parse_state(mediator_variables)
        self.mediation_first_stage_confounders = mediation_first_stage_confounders
        self.mediation_second_stage_confounders = mediation_second_stage_confounders
        self.estimand_type = estimand_type
        self.estimands = estimands
        self.default_backdoor_id = default_backdoor_id
        self.identifier_method = identifier_method
        self.no_directed_path = no_directed_path

    def set_identifier_method(self, identifier_name: str):
        self.identifier_method = identifier_name

    def get_backdoor_variables(self, key: Optional[str] = None):
        """Return a list containing the backdoor variables.

        If the calling estimator method is a backdoor method, return the
        backdoor variables corresponding to its target estimand.
        Otherwise, return the backdoor variables for the default backdoor estimand.
        """
        if key is None:
            if self.identifier_method and self.identifier_method.startswith("backdoor"):
                return self.backdoor_variables[self.identifier_method]
            elif self.backdoor_variables is not None and len(self.backdoor_variables) > 0:
                return self.backdoor_variables[self.default_backdoor_id]
            else:
                return []
        else:
            return self.backdoor_variables[key]

    def set_backdoor_variables(self, bdoor_variables_arr: List, key: Optional[str] = None):
        if key is None:
            key = self.identifier_method
        self.backdoor_variables[key] = bdoor_variables_arr

    def get_frontdoor_variables(self):
        """Return a list containing the frontdoor variables (if present)"""
        return self.frontdoor_variables

    def get_mediator_variables(self):
        """Return a list containing the mediator variables (if present)"""
        return self.mediator_variables

    def get_instrumental_variables(self):
        """Return a list containing the instrumental variables (if present)"""
        return self.instrumental_variables

    def __deepcopy__(self, memo):
        return IdentifiedEstimand(
            self.identifier,  # not deep copied
            copy.deepcopy(self.treatment_variable),
            copy.deepcopy(self.outcome_variable),
            estimand_type=copy.deepcopy(self.estimand_type),
            estimands=copy.deepcopy(self.estimands),
            backdoor_variables=copy.deepcopy(self.backdoor_variables),
            instrumental_variables=copy.deepcopy(self.instrumental_variables),
            frontdoor_variables=copy.deepcopy(self.frontdoor_variables),
            mediator_variables=copy.deepcopy(self.mediator_variables),
            default_backdoor_id=copy.deepcopy(self.default_backdoor_id),
            identifier_method=copy.deepcopy(self.identifier_method),
        )

    def __str__(self, only_target_estimand: bool = False, show_all_backdoor_sets: bool = False):
        if self.no_directed_path:
            s = "No directed path from {0} to {1} in the causal graph.".format(
                self.treatment_variable, self.outcome_variable
            )
            s += "\nCausal effect is zero."
            return s
        s = "Estimand type: {0}\n".format(self.estimand_type)
        i = 1
        has_valid_backdoor = sum("backdoor" in key for key in self.estimands.keys())
        for k, v in self.estimands.items():
            if show_all_backdoor_sets:
                # Do not show backdoor key unless it is the only backdoor set.
                if k == "backdoor" and has_valid_backdoor > 1:
                    continue
            else:
                # Just show the default backdoor set
                if k.startswith("backdoor") and k != "backdoor":
                    continue
            if only_target_estimand and k != self.identifier_method:
                continue
            s += "\n### Estimand : {0}\n".format(i)
            s += "Estimand name: {0}".format(k)
            if k == self.default_backdoor_id:
                s += " (Default)"
            s += "\n"
            if v is None:
                s += "No such variable(s) found!\n"
            else:
                sp_expr_str = sp.pretty(v["estimand"], use_unicode=True)
                s += "Estimand expression:\n{0}\n".format(sp_expr_str)
                j = 1
                for ass_name, ass_str in v["assumptions"].items():
                    s += "Estimand assumption {0}, {1}: {2}\n".format(j, ass_name, ass_str)
                    j += 1
            i += 1
        return s



import copy
from typing import Any, List, Optional, Type, Union

import numpy as np
import pandas as pd

from dowhy.causal_estimator import CausalEstimate, CausalEstimator
from dowhy.causal_estimators.linear_regression_estimator import LinearRegressionEstimator
from dowhy.causal_identifier import EstimandType, IdentifiedEstimand
from dowhy.utils.api import parse_state


class TwoStageRegressionEstimator(CausalEstimator):
    """Compute treatment effect whenever the effect is fully mediated by
    another variable (front-door) or when there is an instrument available.

    Currently only supports a linear model for the effects.

    Supports additional parameters as listed below.

    """

    # First stage statistical model
    DEFAULT_FIRST_STAGE_MODEL = LinearRegressionEstimator
    # Second stage statistical model
    DEFAULT_SECOND_STAGE_MODEL = LinearRegressionEstimator

    def __init__(
        self,
        identified_estimand: IdentifiedEstimand,
        test_significance: bool = False,
        evaluate_effect_strength: bool = False,
        confidence_intervals: bool = False,
        num_null_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_STAT_TEST,
        num_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_CI,
        sample_size_fraction: int = CausalEstimator.DEFAULT_SAMPLE_SIZE_FRACTION,
        confidence_level: float = CausalEstimator.DEFAULT_CONFIDENCE_LEVEL,
        need_conditional_estimates: Union[bool, str] = "auto",
        num_quantiles_to_discretize_cont_cols: int = CausalEstimator.NUM_QUANTILES_TO_DISCRETIZE_CONT_COLS,
        first_stage_model: Optional[Union[CausalEstimator, Type[CausalEstimator]]] = None,
        second_stage_model: Optional[Union[CausalEstimator, Type[CausalEstimator]]] = None,
        **kwargs,
    ):
        """
        :param identified_estimand: probability expression
            representing the target identified estimand to estimate.
        :param test_significance: Binary flag or a string indicating whether to test significance and by which method. All estimators support test_significance="bootstrap" that estimates a p-value for the obtained estimate using the bootstrap method. Individual estimators can override this to support custom testing methods. The bootstrap method supports an optional parameter, num_null_simulations. If False, no testing is done. If True, significance of the estimate is tested using the custom method if available, otherwise by bootstrap.
        :param evaluate_effect_strength: (Experimental) whether to evaluate the strength of effect
        :param confidence_intervals: Binary flag or a string indicating whether the confidence intervals should be computed and which method should be used. All methods support estimation of confidence intervals using the bootstrap method by using the parameter confidence_intervals="bootstrap". The bootstrap method takes in two arguments (num_simulations and sample_size_fraction) that can be optionally specified in the params dictionary. Estimators may also override this to implement their own confidence interval method. If this parameter is False, no confidence intervals are computed. If True, confidence intervals are computed by the estimator's specific method if available, otherwise through bootstrap
        :param num_null_simulations: The number of simulations for testing the
            statistical significance of the estimator
        :param num_simulations: The number of simulations for finding the
            confidence interval (and/or standard error) for a estimate
        :param sample_size_fraction: The size of the sample for the bootstrap
            estimator
        :param confidence_level: The confidence level of the confidence
            interval estimate
        :param need_conditional_estimates: Boolean flag indicating whether
            conditional estimates should be computed. Defaults to True if
            there are effect modifiers in the graph
        :param num_quantiles_to_discretize_cont_cols: The number of quantiles
            into which a numeric effect modifier is split, to enable
            estimation of conditional treatment effect over it.
        :param first_stage_model: First stage estimator to be used. Default is
            linear regression.
        :param second_stage_model: Second stage estimator to be used. Default
            is linear regression.
        :param kwargs: (optional) Additional estimator-specific parameters
        """
        super().__init__(
            identified_estimand=identified_estimand,
            test_significance=test_significance,
            evaluate_effect_strength=evaluate_effect_strength,
            confidence_intervals=confidence_intervals,
            num_null_simulations=num_null_simulations,
            num_simulations=num_simulations,
            sample_size_fraction=sample_size_fraction,
            confidence_level=confidence_level,
            need_conditional_estimates=need_conditional_estimates,
            num_quantiles_to_discretize_cont_cols=num_quantiles_to_discretize_cont_cols,
            first_stage_model=first_stage_model,
            second_stage_model=second_stage_model,
            **kwargs,
        )
        self.logger.info("INFO: Using Two Stage Regression Estimator")
        # Check if the treatment is one-dimensional
        modified_target_estimand = copy.deepcopy(self._target_estimand)
        modified_target_estimand.identifier_method = "backdoor"
        modified_target_estimand.backdoor_variables = self._target_estimand.mediation_first_stage_confounders
        if first_stage_model is not None:
            self._first_stage_model = (
                first_stage_model
                if isinstance(first_stage_model, CausalEstimator)
                else first_stage_model(
                    modified_target_estimand,
                    test_significance=self._significance_test,
                    evaluate_effect_strength=self._effect_strength_eval,
                    confidence_intervals=self._confidence_intervals,
                    **kwargs,
                )
            )
        else:
            self._first_stage_model = self.__class__.DEFAULT_FIRST_STAGE_MODEL(
                modified_target_estimand,
                test_significance=self._significance_test,
                evaluate_effect_strength=self._effect_strength_eval,
                confidence_intervals=self._confidence_intervals,
                **kwargs,
            )
            self.logger.warning("First stage model not provided. Defaulting to sklearn.linear_model.LinearRegression.")

        modified_target_estimand = copy.deepcopy(self._target_estimand)
        modified_target_estimand.identifier_method = "backdoor"
        modified_target_estimand.backdoor_variables = self._target_estimand.mediation_second_stage_confounders
        if second_stage_model is not None:
            self._second_stage_model = (
                second_stage_model
                if isinstance(second_stage_model, CausalEstimator)
                else second_stage_model(
                    modified_target_estimand,
                    test_significance=self._significance_test,
                    evaluate_effect_strength=self._effect_strength_eval,
                    confidence_intervals=self._confidence_intervals,
                    **kwargs,
                )
            )
        else:
            modified_target_estimand = copy.deepcopy(self._target_estimand)
            self._second_stage_model = self.__class__.DEFAULT_SECOND_STAGE_MODEL(
                modified_target_estimand,
                test_significance=self._significance_test,
                evaluate_effect_strength=self._effect_strength_eval,
                confidence_intervals=self._confidence_intervals,
                **kwargs,
            )
            self.logger.warning("Second stage model not provided. Defaulting to backdoor.linear_regression.")

        if self._target_estimand.estimand_type == EstimandType.NONPARAMETRIC_NDE:
            modified_target_estimand.identifier_method = "backdoor"
            modified_target_estimand = copy.deepcopy(self._target_estimand)
            self._second_stage_model_nde = type(self._second_stage_model)(
                modified_target_estimand,
                test_significance=self._significance_test,
                evaluate_effect_strength=self._effect_strength_eval,
                confidence_intervals=self._confidence_intervals,
                **kwargs,
            )

    def fit(
        self,
        data: pd.DataFrame,
        treatment_name: str,
        outcome_name: str,
        effect_modifier_names: Optional[List[str]] = None,
        **_,
    ):
        """
        Fits the estimator with data for effect estimation
        :param data: data frame containing the data
        :param treatment: name of the treatment variable
        :param outcome: name of the outcome variable
        :param iv_instrument_name: Name of the specific instrumental variable
            to be used. Needs to be one of the IVs identified in the
            identification step. Default is to use all the IV variables
            from the identification step.
        :param effect_modifiers: Variables on which to compute separate
                    effects, or return a heterogeneous effect function. Not all
                    methods support this currently.
        """
        self._set_data(data, treatment_name, outcome_name)
        self._set_effect_modifiers(effect_modifier_names)

        if len(self._treatment_name) > 1:
            error_msg = str(self.__class__) + "cannot handle more than one treatment variable"
            raise Exception(error_msg)

        if self._target_estimand.identifier_method == "frontdoor":
            self.logger.debug("Front-door variable used:" + ",".join(self._target_estimand.get_frontdoor_variables()))
            self._frontdoor_variables_names = self._target_estimand.get_frontdoor_variables()

            if self._frontdoor_variables_names:
                self._frontdoor_variables = self._data[self._frontdoor_variables_names]
            else:
                self._frontdoor_variables = None
                error_msg = "No front-door variable present. Two stage regression is not applicable"
                self.logger.error(error_msg)
        elif self._target_estimand.identifier_method == "mediation":
            self.logger.debug("Mediators used:" + ",".join(self._target_estimand.get_mediator_variables()))
            self._mediators_names = self._target_estimand.get_mediator_variables()

            if self._mediators_names:
                self._mediators = self._data[self._mediators_names]
            else:
                self._mediators = None
                error_msg = "No mediator variable present. Two stage regression is not applicable"
                self.logger.error(error_msg)
        elif self._target_estimand.identifier_method == "iv":
            self.logger.debug(
                "Instrumental variables used:" + ",".join(self._target_estimand.get_instrumental_variables())
            )
            self._instrumental_variables_names = self._target_estimand.get_instrumental_variables()

            if self._instrumental_variables_names:
                self._instrumental_variables = self._data[self._instrumental_variables_names]
            else:
                self._instrumental_variables = None
                error_msg = "No instrumental variable present. Two stage regression is not applicable"
                self.logger.error(error_msg)

        if self._target_estimand.identifier_method == "frontdoor":
            self._first_stage_model._target_estimand.outcome_variable = parse_state(self._frontdoor_variables_names)
        elif self._target_estimand.identifier_method == "mediation":
            self._first_stage_model._target_estimand.outcome_variable = parse_state(self._mediators_names)

        self._first_stage_model.fit(
            data,
            treatment_name,
            parse_state(self._first_stage_model._target_estimand.outcome_variable),
            effect_modifier_names=effect_modifier_names,
        )

        if self._target_estimand.identifier_method == "frontdoor":
            self._second_stage_model._target_estimand.treatment_variable = parse_state(self._frontdoor_variables_names)
        elif self._target_estimand.identifier_method == "mediation":
            self._second_stage_model._target_estimand.treatment_variable = parse_state(self._mediators_names)

        self._second_stage_model.fit(
            data,
            parse_state(self._second_stage_model._target_estimand.treatment_variable),
            parse_state(self._outcome_name),  # to convert it to array before passing to causal estimator)
            effect_modifier_names=effect_modifier_names,
        )

        if self._target_estimand.estimand_type == EstimandType.NONPARAMETRIC_NDE:
            self._second_stage_model_nde._target_estimand.identifier_method = "backdoor"
            self._second_stage_model_nde.fit(
                data,
                self._treatment_name,
                parse_state(self._outcome_name),  # to convert it to array before passing to causal estimator)
                effect_modifier_names=effect_modifier_names,
            )

        return self

    def estimate_effect(
        self, data: pd.DataFrame = None, treatment_value: Any = 1, control_value: Any = 0, target_units=None, **_
    ):
        if data is None:
            data = self._data
        self._target_units = target_units
        self._treatment_value = treatment_value
        self._control_value = control_value

        estimate_value = None
        # First stage
        first_stage_estimate = self._first_stage_model.estimate_effect(
            control_value=control_value,
            treatment_value=treatment_value,
            target_units=target_units,
        )

        # Second Stage
        second_stage_estimate = self._second_stage_model.estimate_effect(
            control_value=control_value,
            treatment_value=treatment_value,
            target_units=target_units,
        )
        # Combining the two estimates
        natural_indirect_effect = first_stage_estimate.value * second_stage_estimate.value
        # This same estimate is valid for frontdoor as well as mediation (NIE)
        estimate_value = natural_indirect_effect
        self.symbolic_estimator = self.construct_symbolic_estimator(
            first_stage_estimate.realized_estimand_expr,
            second_stage_estimate.realized_estimand_expr,
            estimand_type=EstimandType.NONPARAMETRIC_NIE,
        )
        if self._target_estimand.estimand_type == EstimandType.NONPARAMETRIC_NDE:

            total_effect_estimate = self._second_stage_model_nde.estimate_effect(
                control_value=control_value, treatment_value=treatment_value, target_units=target_units
            )
            natural_direct_effect = total_effect_estimate.value - natural_indirect_effect
            estimate_value = natural_direct_effect
            self.symbolic_estimator = self.construct_symbolic_estimator(
                first_stage_estimate.realized_estimand_expr,
                second_stage_estimate.realized_estimand_expr,
                total_effect_estimate.realized_estimand_expr,
                estimand_type=self._target_estimand.estimand_type,
            )
        estimate = CausalEstimate(
            estimate=estimate_value,
            control_value=control_value,
            treatment_value=treatment_value,
            target_estimand=self._target_estimand,
            realized_estimand_expr=self.symbolic_estimator,
        )

        estimate.add_estimator(self)
        return estimate

    def build_first_stage_features(self):
        data_df = self._data
        treatment_vals = data_df[self._treatment_name]
        if len(self._observed_common_causes_names) > 0:
            observed_common_causes_vals = data_df[self._observed_common_causes_names]
            observed_common_causes_vals = pd.get_dummies(observed_common_causes_vals, drop_first=True)
        if self._effect_modifier_names:
            effect_modifiers_vals = data_df[self._effect_modifier_names]
            effect_modifiers_vals = pd.get_dummies(effect_modifiers_vals, drop_first=True)
        if type(treatment_vals) is not np.ndarray:
            treatment_vals = treatment_vals.to_numpy()
        if treatment_vals.shape[0] != data_df.shape[0]:
            raise ValueError("Provided treatment values and dataframe should have the same length.")
        # Bulding the feature matrix
        n_samples = treatment_vals.shape[0]
        self.logger.debug("Number of samples" + str(n_samples) + str(len(self._treatment_name)))
        treatment_2d = treatment_vals.reshape((n_samples, len(self._treatment_name)))
        if len(self._observed_common_causes_names) > 0:
            features = np.concatenate((treatment_2d, observed_common_causes_vals), axis=1)
        else:
            features = treatment_2d
        if self._effect_modifier_names:
            for i in range(treatment_2d.shape[1]):
                curr_treatment = treatment_2d[:, i]
                new_features = curr_treatment[:, np.newaxis] * effect_modifiers_vals.to_numpy()
                features = np.concatenate((features, new_features), axis=1)
        features = features.astype(
            float, copy=False
        )  # converting to float in case of binary treatment and no other variables
        # features = sm.add_constant(features, has_constant='add') # to add an intercept term
        return features

    def construct_symbolic_estimator(
        self, first_stage_symbolic, second_stage_symbolic, total_effect_symbolic=None, estimand_type=None
    ):
        nie_symbolic = "(" + first_stage_symbolic + ")*(" + second_stage_symbolic + ")"
        if estimand_type == EstimandType.NONPARAMETRIC_NIE:
            return nie_symbolic
        elif estimand_type == EstimandType.NONPARAMETRIC_NDE:
            return "(" + total_effect_symbolic + ") - (" + nie_symbolic + ")"



import itertools
from typing import Any, List, Optional, Union

import pandas as pd
import statsmodels.api as sm

from dowhy.causal_estimator import CausalEstimator
from dowhy.causal_estimators.regression_estimator import RegressionEstimator
from dowhy.causal_identifier import IdentifiedEstimand


class GeneralizedLinearModelEstimator(RegressionEstimator):
    """Compute effect of treatment using a generalized linear model such as logistic regression.

    Implementation uses statsmodels.api.GLM.
    Needs an additional parameter, "glm_family" to be specified in method_params. The value of this parameter can be any valid statsmodels.api families object. For example, to use logistic regression, specify "glm_family" as statsmodels.api.families.Binomial().

    """

    def __init__(
        self,
        identified_estimand: IdentifiedEstimand,
        test_significance: bool = False,
        evaluate_effect_strength: bool = False,
        confidence_intervals: bool = False,
        num_null_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_STAT_TEST,
        num_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_CI,
        sample_size_fraction: int = CausalEstimator.DEFAULT_SAMPLE_SIZE_FRACTION,
        confidence_level: float = CausalEstimator.DEFAULT_CONFIDENCE_LEVEL,
        need_conditional_estimates: Union[bool, str] = "auto",
        num_quantiles_to_discretize_cont_cols: int = CausalEstimator.NUM_QUANTILES_TO_DISCRETIZE_CONT_COLS,
        glm_family: Optional[Any] = None,
        predict_score: bool = True,
        **kwargs,
    ):
        """For a list of args and kwargs, see documentation for
        :class:`~dowhy.causal_estimator.CausalEstimator`.

        :param identified_estimand: probability expression
            representing the target identified estimand to estimate.
        :param test_significance: Binary flag or a string indicating whether to test significance and by which method. All estimators support test_significance="bootstrap" that estimates a p-value for the obtained estimate using the bootstrap method. Individual estimators can override this to support custom testing methods. The bootstrap method supports an optional parameter, num_null_simulations. If False, no testing is done. If True, significance of the estimate is tested using the custom method if available, otherwise by bootstrap.
        :param evaluate_effect_strength: (Experimental) whether to evaluate the strength of effect
        :param confidence_intervals: Binary flag or a string indicating whether the confidence intervals should be computed and which method should be used. All methods support estimation of confidence intervals using the bootstrap method by using the parameter confidence_intervals="bootstrap". The bootstrap method takes in two arguments (num_simulations and sample_size_fraction) that can be optionally specified in the params dictionary. Estimators may also override this to implement their own confidence interval method. If this parameter is False, no confidence intervals are computed. If True, confidence intervals are computed by the estimator's specific method if available, otherwise through bootstrap
        :param num_null_simulations: The number of simulations for testing the
            statistical significance of the estimator
        :param num_simulations: The number of simulations for finding the
            confidence interval (and/or standard error) for a estimate
        :param sample_size_fraction: The size of the sample for the bootstrap
            estimator
        :param confidence_level: The confidence level of the confidence
            interval estimate
        :param need_conditional_estimates: Boolean flag indicating whether
            conditional estimates should be computed. Defaults to True if
            there are effect modifiers in the graph
        :param num_quantiles_to_discretize_cont_cols: The number of quantiles
            into which a numeric effect modifier is split, to enable
            estimation of conditional treatment effect over it.
        :param glm_family: statsmodels family for the generalized linear model.
            For example, use statsmodels.api.families.Binomial() for logistic
            regression or statsmodels.api.families.Poisson() for count data.
        :param predict_score: For models that have a binary output, whether
            to output the model's score or the binary output based on the score.
        :param kwargs: (optional) Additional estimator-specific parameters


        """
        super().__init__(
            identified_estimand=identified_estimand,
            test_significance=test_significance,
            evaluate_effect_strength=evaluate_effect_strength,
            confidence_intervals=confidence_intervals,
            num_null_simulations=num_null_simulations,
            num_simulations=num_simulations,
            sample_size_fraction=sample_size_fraction,
            confidence_level=confidence_level,
            need_conditional_estimates=need_conditional_estimates,
            num_quantiles_to_discretize_cont_cols=num_quantiles_to_discretize_cont_cols,
            glm_family=glm_family,
            predict_score=predict_score,
            **kwargs,
        )
        self.logger.info("INFO: Using Generalized Linear Model Estimator")
        if glm_family is not None:
            self.family = glm_family
        else:
            raise ValueError(
                "Need to specify the family for the generalized linear model. Provide a 'glm_family' parameter in method_params, such as statsmodels.api.families.Binomial() for logistic regression."
            )
        self.predict_score = predict_score

    def fit(
        self,
        data: pd.DataFrame,
        treatment_name: str,
        outcome_name: str,
        effect_modifier_names: Optional[List[str]] = None,
    ):
        """
        Fits the estimator with data for effect estimation
        :param data: data frame containing the data
        :param treatment: name of the treatment variable
        :param outcome: name of the outcome variable
        :param effect_modifiers: Variables on which to compute separate
                    effects, or return a heterogeneous effect function. Not all
                    methods support this currently.
        """
        return super().fit(data, treatment_name, outcome_name, effect_modifier_names=effect_modifier_names)

    def _build_model(self):
        features = self._build_features()
        model = sm.GLM(self._outcome, features, family=self.family).fit()
        return (features, model)

    def predict_fn(self, model, features):
        # Checking if Y is binary
        outcome_values = self._data[self._outcome_name].astype(int).unique()
        outcome_is_binary = all([v in [0, 1] for v in outcome_values])

        if outcome_is_binary:
            if self.predict_score:
                return model.predict(features)
            else:
                return (model.predict(features) > 0.5).astype(int)
        else:
            return model.predict(features)

    def construct_symbolic_estimator(self, estimand):
        expr = "b: " + ",".join(estimand.outcome_variable) + "~" + "Sigmoid("
        var_list = estimand.treatment_variable + estimand.get_backdoor_variables()
        expr += "+".join(var_list)
        if self._effect_modifier_names:
            interaction_terms = [
                "{0}*{1}".format(x[0], x[1])
                for x in itertools.product(estimand.treatment_variable, self._effect_modifier_names)
            ]
            expr += "+" + "+".join(interaction_terms)
        expr += ")"
        return expr



import networkx as nx
import numpy as np
import pandas as pd
from flaky import flaky
from pytest import approx

from dowhy.gcm import (
    AdditiveNoiseModel,
    InverseDensityScorer,
    InvertibleStructuralCausalModel,
    MedianCDFQuantileScorer,
    PredictionModel,
    attribute_anomalies,
    auto,
    fit,
)
from dowhy.gcm.anomaly import _relative_frequency, attribute_anomaly_scores
from dowhy.gcm.density_estimators import GaussianMixtureDensityEstimator


@flaky(max_runs=3)
def test_given_simple_gaussian_data_when_attribute_anomaly_scores_with_it_score_then_returns_qualitatively_correct_results():
    Z = np.random.normal(0, 1, 5000)
    X0 = Z + np.random.normal(0, 1, 5000)
    X1 = Z + np.random.normal(0, 1, 5000)
    X2 = Z + np.random.normal(0, 1, 5000)
    X3 = Z + np.random.normal(0, 1, 5000)

    original_observations = np.column_stack([X0, X1, X2, X3])

    # Defining an anomaly scorer that handles multidimensional inputs.
    density_estimator = GaussianMixtureDensityEstimator()
    anomaly_scorer = InverseDensityScorer(density_estimator)
    anomaly_scorer.fit(original_observations)

    # Seeing that the expectation of the noise in all nodes is 0, we introduce anomalies by setting some of them to 3.
    anomaly_samples = np.array([[3, 0, 0, 0], [0, 3, 0, 0], [0, 0, 3, 0], [0, 3, 3, 0], [3, 0, 3, 3]])

    contributions = attribute_anomaly_scores(anomaly_samples, original_observations, anomaly_scorer.score, False)

    # In the first sample, only the first variable is anomalous. Therefore, it should have the highest contribution
    # and it should be "significantly" higher than the contribution of the other ones (here, we just arbitrarily say
    # it should be 3x higher. Due to the confounding factor Z, the reconstructed noise variables are pairwise dependent,
    # which is a violation of our causal sufficiency assumption. However, a confounder is included here to demonstrate
    # some robustness. Note that due to this and stochastic behaviour of the density estimator, it is
    # not possible to analytically compute expected results. Therefore, we rather look at the relations here.
    assert np.argmax(contributions[0]) == 0
    assert np.all(contributions[0][0] > contributions[0][1:] * 3)

    # Same idea for the second sample, but here, it is the second variable that is anomalous.
    assert np.argmax(contributions[1]) == 1
    assert np.all(contributions[1][1] > contributions[1][0] * 3)
    assert np.all(contributions[1][1] > contributions[1][2:] * 3)

    assert np.argmax(contributions[2]) == 2
    assert np.all(contributions[2][2] > contributions[2][0:2] * 3)
    assert np.all(contributions[2][2] > contributions[2][3:] * 3)

    # In the fourth sample, there are 2 anomalous variables. Therefore, the contribution of these 2 variables should be
    # "significantly" higher than the contribution of the other variables. The contribution of both anomalous variables
    # should be equal (approximately).
    assert np.all(contributions[3][1] > contributions[3][0] * 3)
    assert np.all(contributions[3][2] > contributions[3][0] * 3)
    assert np.all(contributions[3][1] > contributions[3][3] * 3)
    assert np.all(contributions[3][2] > contributions[3][3] * 3)
    assert contributions[3][1] == approx(contributions[3][2], abs=0.5)

    assert np.all(contributions[4][0] > contributions[4][1] * 3)
    assert np.all(contributions[4][2] > contributions[4][1] * 3)
    assert np.all(contributions[4][3] > contributions[4][1] * 3)
    assert contributions[4][0] == approx(contributions[4][2], abs=0.5)
    assert contributions[4][2] == approx(contributions[4][3], abs=0.5)


@flaky(max_runs=3)
def test_given_simple_gaussian_data_when_attribute_anomaly_scores_with_feature_relevance_then_returns_qualitatively_correct_results():
    X0 = np.random.normal(0, 1, 5000)
    X1 = np.random.normal(0, 1, 5000)
    X2 = np.random.normal(0, 1, 5000)
    X3 = np.random.normal(0, 1, 5000)

    original_observations = np.column_stack([X0, X1, X2, X3])

    # Defining an anomaly scorer that handles multidimensional inputs.
    density_estimator = GaussianMixtureDensityEstimator(num_components=5)
    anomaly_scorer = InverseDensityScorer(density_estimator)
    anomaly_scorer.fit(original_observations)
    expectation_of_score = np.mean(anomaly_scorer.score(original_observations))

    # Seeing that the expectation of the noise in all nodes is 0, we introduce anomalies by setting some of them to 3.
    anomaly_samples = np.array([[3, 0, 0, 0], [0, 3, 3, 0], [3, 0, 3, 3]])

    contributions = attribute_anomaly_scores(anomaly_samples, original_observations, anomaly_scorer.score, True)

    assert np.argmax(contributions[0]) == 0  # The biggest contribution to the score should be for the first variable.
    assert contributions[0][0] > 0  # The contribution should be positive (anomalous variable).
    assert np.all(contributions[0][1:] < 0)  # Contributions should be negative (non-anomalous variables), since they
    # reduce the score.
    # The contributions should add up to g(x) - E[g(X)]
    assert np.sum(contributions[0]) == approx(
        anomaly_scorer.score(anomaly_samples[0].reshape(1, -1)) - expectation_of_score
    )

    assert contributions[1][1] > 0  # The contribution should be positive (anomalous variable).
    assert contributions[1][2] > 0  # The contribution should be positive (anomalous variable).
    assert contributions[1][0] < 0  # The contribution should be negative (non-anomalous variable).
    assert contributions[1][3] < 0  # The contribution should be negative (non-anomalous variable).
    # The contributions should add up to g(x) - E[g(X)]
    assert np.sum(contributions[1]) == approx(
        anomaly_scorer.score(anomaly_samples[1].reshape(1, -1)) - expectation_of_score
    )

    assert contributions[2][0] > 0  # The contribution should be positive (anomalous variable).
    assert contributions[2][2] > 0  # The contribution should be positive (anomalous variable).
    assert contributions[2][3] > 0  # The contribution should be positive (anomalous variable).
    assert contributions[2][1] < 0  # The contribution should be negative (non-anomalous variable).
    # The contributions should add up to g(x) - E[g(X)]
    assert np.sum(contributions[2]) == approx(
        anomaly_scorer.score(anomaly_samples[2].reshape(1, -1)) - expectation_of_score
    )


@flaky(max_runs=3)
def test_given_simple_causal_chain_with_linear_relationships_when_attribute_anomaly_scores_with_it_score_then_returns_qualitatively_correct_results():
    num_training_samples = 5000
    X0 = np.random.normal(0, 1, num_training_samples)
    X1 = X0 + np.random.normal(0, 1, num_training_samples)
    X2 = X1 + np.random.normal(0, 1, num_training_samples)
    X3 = X2 + np.random.normal(0, 1, num_training_samples)
    training_data = pd.DataFrame({"X0": X0, "X1": X1, "X2": X2, "X3": X3})

    causal_model = InvertibleStructuralCausalModel(nx.DiGraph([("X0", "X1"), ("X1", "X2"), ("X2", "X3")]))
    auto.assign_causal_mechanisms(causal_model, training_data, auto.AssignmentQuality.GOOD)

    fit(causal_model, training_data)

    # Three examples:
    # 1. X1 is the root cause (+ 10 to the noise)
    # 2. X0 is the root cause (+ 10 to the noise)
    # 3. X0 and X3 are both root causes (+ 10 to both noises)
    anomaly_data = pd.DataFrame(
        {
            "X0": np.array([0, 10, 10]),
            "X1": np.array([10, 10, 10]),
            "X2": np.array([10, 10, 10]),
            "X3": np.array([10, 10, 20]),
        }
    )

    scores = attribute_anomalies(
        causal_model,
        "X3",
        anomaly_data,
        anomaly_scorer=MedianCDFQuantileScorer(),
        num_distribution_samples=num_training_samples,
        attribute_mean_deviation=False,
    )

    def total_anomaly_score(training_data, anomaly_sample):
        distribution_samples = training_data["X3"].to_numpy()
        anomaly_scorer = MedianCDFQuantileScorer()
        anomaly_scorer.fit(distribution_samples)
        return -np.log(
            _relative_frequency(anomaly_scorer.score(distribution_samples) >= anomaly_scorer.score(anomaly_sample))
        )

    assert scores["X0"][0] < 0.5  # Not anomalous.
    assert scores["X0"][1] > 8  # Anomalous, but the only root cause.
    assert scores["X0"][2] > 4  # Anomalous together with X1.
    assert scores["X1"][0] > 8  # Anomalous, but the only root cause.
    assert scores["X1"][1] < 0.5  # Not anomalous given upstream nodes.
    assert scores["X1"][2] < 0.5  # Not anomalous given upstream nodes.
    assert np.all(scores["X2"] < 0.5)  # Not anomalous given upstream nodes.
    assert np.all(scores["X3"][:2] < 0.5)  # Not anomalous given upstream nodes.
    assert scores["X3"][2] > 4  # Anomalous together with X0.

    # The sum of the scores should add up to the anomaly score of the target (here, X3).
    assert scores["X0"][0] + scores["X1"][0] + scores["X2"][0] + scores["X3"][0] == approx(
        total_anomaly_score(training_data, anomaly_data["X3"].to_numpy()[0]), abs=0.001
    )
    assert scores["X0"][1] + scores["X1"][1] + scores["X2"][1] + scores["X3"][1] == approx(
        total_anomaly_score(training_data, anomaly_data["X3"].to_numpy()[1]), abs=0.001
    )
    assert scores["X0"][2] + scores["X1"][2] + scores["X2"][2] + scores["X3"][2] == approx(
        total_anomaly_score(training_data, anomaly_data["X3"].to_numpy()[2]), abs=0.001
    )


@flaky(max_runs=3)
def test_given_simple_causal_chain_with_linear_relationships_when_attribute_anomaly_scores_with_feature_relevance_then_returns_qualitatively_correct_results():
    num_training_samples = 5000
    X0 = np.random.normal(0, 1, num_training_samples)
    X1 = X0 + np.random.normal(0, 1, num_training_samples)
    X2 = X1 + np.random.normal(0, 1, num_training_samples)
    X3 = X2 + np.random.normal(0, 1, num_training_samples)
    training_data = pd.DataFrame({"X0": X0, "X1": X1, "X2": X2, "X3": X3})

    causal_model = InvertibleStructuralCausalModel(nx.DiGraph([("X0", "X1"), ("X1", "X2"), ("X2", "X3")]))
    auto.assign_causal_mechanisms(causal_model, training_data, auto.AssignmentQuality.GOOD)

    fit(causal_model, training_data)

    # Three examples:
    # 1. X1 is the root cause (+ 10 to the noise)
    # 2. X0 is the root cause (+ 10 to the noise)
    # 3. X0 and X3 are both root causes (+ 10 to both noises)
    anomaly_data = pd.DataFrame(
        {
            "X0": np.array([0, 10, 10]),
            "X1": np.array([10, 10, 10]),
            "X2": np.array([10, 10, 10]),
            "X3": np.array([10, 10, 20]),
        }
    )

    scores = attribute_anomalies(
        causal_model,
        "X3",
        anomaly_data,
        anomaly_scorer=MedianCDFQuantileScorer(),
        num_distribution_samples=num_training_samples,
        attribute_mean_deviation=True,
    )

    def total_anomaly_score(training_data, anomaly_sample):
        distribution_samples = training_data["X3"].to_numpy()
        anomaly_scorer = MedianCDFQuantileScorer()
        anomaly_scorer.fit(distribution_samples)
        return anomaly_scorer.score(anomaly_sample) - np.mean(anomaly_scorer.score(distribution_samples))

    assert scores["X0"][0] == approx(0, abs=0.1)  # Not anomalous.
    assert scores["X0"][1] > 0.4  # Anomalous, but the only root cause.
    assert scores["X0"][2] > 0.2  # Anomalous together with X1.
    assert scores["X1"][0] > 0.4  # Anomalous, but the only root cause.
    assert scores["X1"][1] == approx(0, abs=0.1)  # Not anomalous given upstream nodes.
    assert scores["X1"][2] == approx(0, abs=0.1)  # Not anomalous given upstream nodes.
    assert scores["X2"] == approx(np.array([0, 0, 0]), abs=0.1)  # Not anomalous given upstream nodes.
    assert scores["X3"][:2] == approx(np.array([0, 0]), abs=0.1)  # Not anomalous given upstream nodes.
    assert scores["X3"][2] > 0.2  # Anomalous together with X0.

    # The sum of the scores should add up to the anomaly score of the target (here, X3).
    assert scores["X0"][0] + scores["X1"][0] + scores["X2"][0] + scores["X3"][0] == approx(
        total_anomaly_score(training_data, anomaly_data["X3"].to_numpy()[0]), abs=0.001
    )
    assert scores["X0"][1] + scores["X1"][1] + scores["X2"][1] + scores["X3"][1] == approx(
        total_anomaly_score(training_data, anomaly_data["X3"].to_numpy()[1]), abs=0.001
    )
    assert scores["X0"][2] + scores["X1"][2] + scores["X2"][2] + scores["X3"][2] == approx(
        total_anomaly_score(training_data, anomaly_data["X3"].to_numpy()[2]), abs=0.001
    )


@flaky(max_runs=3)
def test_given_non_trivial_graph_with_nonlinear_relationships_when_attribute_anomaly_scores_with_it_score_then_returns_qualitatively_correct_results():
    num_training_samples = 10000

    # Defining ground truth model to avoid SCM learning issues and, hence, to focus on the anomaly attribution
    # algorithm.
    class MyNonLinearModel(PredictionModel):
        def fit(self, X: np.ndarray, Y: np.ndarray) -> None:
            pass

        def predict(self, X: np.ndarray) -> np.ndarray:
            return (np.sum(X, axis=1) ** 2).reshape(-1, 1)

        def clone(self):
            pass

    X0 = np.random.normal(0, 1, num_training_samples)
    X1 = MyNonLinearModel().predict(X0.reshape(-1, 1)) + np.random.normal(0, 1, (num_training_samples, 1))  # X0**2 + N
    X2 = np.random.uniform(-5, 5, num_training_samples)
    X3 = MyNonLinearModel().predict(np.column_stack([X1, X2])) + np.random.normal(
        0, 1, (num_training_samples, 1)
    )  # (X1 + X2)**2 + N

    training_data = pd.DataFrame({"X0": X0, "X1": X1.reshape(-1), "X2": X2, "X3": X3.reshape(-1)})
    causal_model = InvertibleStructuralCausalModel(nx.DiGraph([("X0", "X1"), ("X1", "X3"), ("X2", "X3")]))
    causal_model.set_causal_mechanism("X1", AdditiveNoiseModel(MyNonLinearModel()))
    causal_model.set_causal_mechanism("X3", AdditiveNoiseModel(MyNonLinearModel()))
    auto.assign_causal_mechanisms(causal_model, training_data, auto.AssignmentQuality.GOOD, override_models=False)

    fit(causal_model, training_data)

    anomaly_data = pd.DataFrame(
        {
            "X0": np.array([10, 10, 0]),
            "X1": np.array([100, 110, 0]),
            "X2": np.array([0, 0, 10]),
            "X3": np.array([10000, 12100, 110]),
        }
    )

    scores = attribute_anomalies(
        causal_model,
        "X3",
        anomaly_data,
        anomaly_scorer=MedianCDFQuantileScorer(),
        num_distribution_samples=num_training_samples,
        attribute_mean_deviation=False,
    )

    def total_anomaly_score(training_data, anomaly_sample):
        distribution_samples = training_data["X3"].to_numpy()
        anomaly_scorer = MedianCDFQuantileScorer()
        anomaly_scorer.fit(distribution_samples)
        return -np.log(
            _relative_frequency(anomaly_scorer.score(distribution_samples) >= anomaly_scorer.score(anomaly_sample))
        )

    # 1. X0 is the root cause (+ 10 to the noise)
    # 2. X0 and X1 are the root causes (+ 10 to both noise)
    # 3. X2 and X3 are both root causes (+ 10 to both noises)
    assert scores["X0"][0] > 8  # Anomalous, but the only root cause.
    assert scores["X0"][1] > 4  # Anomalous, but together with X1.
    assert scores["X0"][2] < 1  # Not anomalous.

    assert scores["X1"][0] < 1  # Not anomalous given upstream nodes.
    assert scores["X1"][1] > 1  # Anomalous, but together with X0.
    assert scores["X1"][2] < 1  # Not anomalous given upstream nodes.

    assert scores["X2"][0] < 1  # Not anomalous.
    assert scores["X2"][1] < 1  # Not anomalous.
    assert scores["X2"][2] > 1  # Anomalous, but together with X3.

    assert scores["X3"][0] < 1  # Not anomalous given upstream nodes.
    assert scores["X3"][1] < 1  # Not anomalous given upstream nodes.
    assert scores["X3"][2] > 1  # Anomalous, but together with X1.

    # The sum of the scores should add up to the anomaly score of the target (here, X3).
    assert scores["X0"][0] + scores["X1"][0] + scores["X2"][0] + scores["X3"][0] == approx(
        total_anomaly_score(training_data, anomaly_data["X3"].to_numpy()[0]), abs=0.4
    )
    assert scores["X0"][1] + scores["X1"][1] + scores["X2"][1] + scores["X3"][1] == approx(
        total_anomaly_score(training_data, anomaly_data["X3"].to_numpy()[1]), abs=0.4
    )
    assert scores["X0"][2] + scores["X1"][2] + scores["X2"][2] + scores["X3"][2] == approx(
        total_anomaly_score(training_data, anomaly_data["X3"].to_numpy()[2]), abs=0.6
    )


def test_relative_frequency():
    assert np.abs(_relative_frequency(np.array([True, True, False, True])) - 4 / 5) < 0.1



import pytest

from dowhy.data_transformer import DimensionalityReducer


class MockReducer(DimensionalityReducer):
    pass


def test_dimensionality_reducer_placeholder_methods():
    reducer = MockReducer(None, None)
    with pytest.raises(NotImplementedError):
        reducer.reduce()



"""This module provides functionality for fitting probabilistic causal models and drawing samples from them.

Functions in this module should be considered experimental, meaning there might be breaking API changes in the future.
"""

from typing import Any

import networkx as nx
import numpy as np
import pandas as pd
from tqdm import tqdm

from dowhy.gcm import config
from dowhy.gcm.cms import ProbabilisticCausalModel
from dowhy.gcm.graph import (
    PARENTS_DURING_FIT,
    get_ordered_predecessors,
    is_root_node,
    validate_causal_dag,
    validate_causal_model_assignment,
)


def fit(causal_model: ProbabilisticCausalModel, data: pd.DataFrame):
    """Learns generative causal models of nodes in the causal graph from data.

    :param causal_model: The causal model containing the mechanisms that will be fitted.
    :param data: Observations of nodes in the causal model.
    """
    progress_bar = tqdm(
        causal_model.graph.nodes,
        desc="Fitting causal models",
        position=0,
        leave=True,
        disable=not config.show_progress_bars,
    )
    for node in progress_bar:
        if node not in data:
            raise RuntimeError(
                "Could not find data for node %s in the given training data! There should be a column "
                "containing samples for node %s." % (node, node)
            )

        progress_bar.set_description("Fitting causal mechanism of node %s" % node)

        fit_causal_model_of_target(causal_model, node, data)


def fit_causal_model_of_target(
    causal_model: ProbabilisticCausalModel, target_node: Any, training_data: pd.DataFrame
) -> None:
    """Fits only the causal mechanism of the given target node based on the training data.

    :param causal_model: The causal model containing the target node.
    :param target_node: Target node for which the mechanism is fitted.
    :param training_data: Training data for fitting the causal mechanism.
    :return: None
    """
    validate_causal_model_assignment(causal_model.graph, target_node)

    if is_root_node(causal_model.graph, target_node):
        causal_model.causal_mechanism(target_node).fit(X=training_data[target_node].to_numpy())
    else:
        causal_model.causal_mechanism(target_node).fit(
            X=training_data[get_ordered_predecessors(causal_model.graph, target_node)].to_numpy(),
            Y=training_data[target_node].to_numpy(),
        )

    # To be able to validate that the graph structure did not change between fitting and causal query, we store the
    # parents of a node during fit. That way, before sampling, we can verify the parents are still the same. While
    # this would automatically fail when the number of parents is different, there are other more subtle cases,
    # where the number is still the same, but it's different parents, and therefore different data. That would yield
    # wrong results, but would not fail.
    causal_model.graph.nodes[target_node][PARENTS_DURING_FIT] = get_ordered_predecessors(
        causal_model.graph, target_node
    )


def draw_samples(causal_model: ProbabilisticCausalModel, num_samples: int) -> pd.DataFrame:
    """Draws new joint samples from the given graphical causal model. This is done by first generating random samples
    from root nodes and then propagating causal downstream effects through the graph.

    :param causal_model: New samples are generated based on the given causal model.
    :param num_samples: Number of samples to draw.
    :return: A pandas data frame where columns correspond to the nodes in the graph and rows to the drawn joint samples.
    """
    validate_causal_dag(causal_model.graph)

    sorted_nodes = list(nx.topological_sort(causal_model.graph))
    drawn_samples = pd.DataFrame(np.empty((num_samples, len(sorted_nodes))), columns=sorted_nodes)

    for node in sorted_nodes:
        causal_mechanism = causal_model.causal_mechanism(node)

        if is_root_node(causal_model.graph, node):
            drawn_samples[node] = causal_mechanism.draw_samples(num_samples)
        else:
            drawn_samples[node] = causal_mechanism.draw_samples(_parent_samples_of(node, causal_model, drawn_samples))

    return drawn_samples


def _parent_samples_of(node: Any, scm: ProbabilisticCausalModel, samples: pd.DataFrame) -> np.ndarray:
    return samples[get_ordered_predecessors(scm.graph, node)].to_numpy()



from .generalised_cov_measure import generalised_cov_based
from .kernel import approx_kernel_based, kernel_based
from .regression import regression_based


def independence_test(X, Y, conditioned_on=None, method="kernel", **kwargs):
    """Performs a (conditional) independence test.
    Three methods for (conditional) independence test are supported at the moment:

    * `kernel`: Kernel-based (conditional) independence test.

      * K. Zhang, J. Peters, D. Janzing, B. Schölkopf. *Kernel-based Conditional Independence Test and Application in Causal Discovery*. UAI'11, Pages 804–813, 2011.
      * A. Gretton, K. Fukumizu, C.-H. Teo, L. Song, B. Schölkopf, A. Smola. *A Kernel Statistical Test of Independence*. NIPS 21, 2007.
      Here, we utilize the implementations of the https://github.com/cmu-phil/causal-learn package.

    * `approx_kernel`: Approximate kernel-based (conditional) independence test.

      * E. Strobl, K. Zhang, S. Visweswaran. *Approximate kernel-based conditional independence tests for fast non-parametric causal discovery*. Journal of Causal Inference, 2019.

    * `regression`: Regression based (conditional) independence test using a f-test. See :func:`~dowhy.gcm.regression_based` for more details.

    * `gcm`: (Conditional) independence test based on the Generalised Covariance Measure. See :func:`~dowhy.gcm.generalised_cov_based` for more details.

        * R. D. Shah and J Peters. *The hardness of conditional independence testing and the generalised covariance measure*, The Annals of Statistics 48(3), 2018

    :param X: Observations of X.
    :param Y: Observations of Y.
    :param conditioned_on: Observations of conditioning variable if we want to perform a conditional independence test. By default, independence test is carried out.
    :param method: Method for conditional independence test. The choices are:
                   `kernel` (default): :func:`~dowhy.gcm.kernel_based` (conditional) independence test.
                   `approx_kernel`: :func:`~dowhy.gcm.approx_kernel_based` (conditional) independence test.
                   `regression`: :func:`~dowhy.gcm.regression_based` (conditional) independence test.
                   `gcm`: :func:`~dowhy.gcm.generalised_cov_based` (conditional) independence test.
                   For more information about these methods, see above.
    :return:  p-value of the (conditional) independence test. (Conditional) Independence is the null hypothesis.
    """
    if method == "kernel":
        return kernel_based(X, Y, Z=conditioned_on, **kwargs)
    elif method == "approx_kernel":
        return approx_kernel_based(X, Y, Z=conditioned_on, **kwargs)
    elif method == "regression":
        return regression_based(X, Y, Z=conditioned_on, **kwargs)
    elif method == "gcm":
        return generalised_cov_based(X, Y, Z=conditioned_on, **kwargs)
    else:
        raise ValueError(f'Invalid method "{method}"')



"""This module defines the fundamental classes for graphical causal models (GCMs).

Classes in this module should be considered experimental, meaning there might be breaking API changes in the future.
"""

from typing import Any, Callable, Optional, Union

import networkx as nx

from dowhy.gcm.graph import (
    CAUSAL_MECHANISM,
    ConditionalStochasticModel,
    DirectedGraph,
    FunctionalCausalModel,
    InvertibleFunctionalCausalModel,
    StochasticModel,
    clone_causal_models,
)


class ProbabilisticCausalModel:
    """Represents a probabilistic graphical causal model, i.e. it combines a graphical representation of causal
    causal relationships and corresponding causal mechanism for each node describing the data generation process. The
    causal mechanisms can be any general stochastic models."""

    def __init__(
        self, graph: Optional[DirectedGraph] = None, graph_copier: Callable[[DirectedGraph], DirectedGraph] = nx.DiGraph
    ):
        """
        :param graph: Optional graph object to be used as causal graph.
        :param graph_copier: Optional function that can copy a causal graph. Defaults to a networkx.DiGraph
                             constructor.
        """
        if graph is None:
            graph = nx.DiGraph()
        self.graph = graph
        self.graph_copier = graph_copier

    def set_causal_mechanism(self, node: Any, mechanism: Union[StochasticModel, ConditionalStochasticModel]) -> None:
        """Assigns the generative causal model of node in the causal graph.

        :param node: Target node whose causal model is to be assigned.
        :param mechanism: Causal mechanism to be assigned. A root node must be a
                          :class:`~dowhy.gcm.graph.StochasticModel`, whereas a non-root node must be a
                          :class:`~dowhy.gcm.graph.ConditionalStochasticModel`.
        """
        if node not in self.graph.nodes:
            raise ValueError("Node %s can not be found in the given graph!" % node)
        self.graph.nodes[node][CAUSAL_MECHANISM] = mechanism

    def causal_mechanism(self, node: Any) -> Union[StochasticModel, ConditionalStochasticModel]:
        """Returns the generative causal model of node in the causal graph.

        :param node: Target node whose causal model is to be assigned.
        :returns: The causal mechanism for this node. A root node is of type
                  :class:`~dowhy.gcm.graph.StochasticModel`, whereas a non-root node is of type
                  :class:`~dowhy.gcm.graph.ConditionalStochasticModel`.
        """
        return self.graph.nodes[node][CAUSAL_MECHANISM]

    def clone(self):
        """Clones the causal model, but keeps causal mechanisms untrained."""
        graph_copy = self.graph_copier(self.graph)
        clone_causal_models(self.graph, graph_copy)
        return self.__class__(graph_copy)


class StructuralCausalModel(ProbabilisticCausalModel):
    """Represents a structural causal model (SCM), as required e.g. by
    :func:`~dowhy.gcm.whatif.counterfactual_samples`. As compared to a :class:`~dowhy.gcm.cms.ProbabilisticCausalModel`,
    an SCM describes the data generation process in non-root nodes by functional causal models.
    """

    def set_causal_mechanism(self, node: Any, mechanism: Union[StochasticModel, FunctionalCausalModel]) -> None:
        super().set_causal_mechanism(node, mechanism)

    def causal_mechanism(self, node: Any) -> Union[StochasticModel, FunctionalCausalModel]:
        return super().causal_mechanism(node)


class InvertibleStructuralCausalModel(StructuralCausalModel):
    """Represents an invertible structural graphical causal model, as required e.g. by
    :func:`~dowhy.gcm.whatif.counterfactual_samples`. This is a subclass of
    :class:`~dowhy.gcm.cms.StructuralCausalModel` and has further restrictions on the class of causal mechanisms.
    Here, the mechanisms of non-root nodes need to be invertible with respect to the noise,
    such as :class:`~dowhy.gcm.fcms.PostNonlinearModel`.
    """

    def set_causal_mechanism(
        self, target_node: Any, mechanism: Union[StochasticModel, InvertibleFunctionalCausalModel]
    ) -> None:
        super().set_causal_mechanism(target_node, mechanism)

    def causal_mechanism(self, node: Any) -> Union[StochasticModel, InvertibleFunctionalCausalModel]:
        return super().causal_mechanism(node)



from importlib import import_module

import networkx as nx
import numpy as np

from dowhy.graph_learners import GraphLearner
from dowhy.utils.graph_operations import *


class GES(GraphLearner):
    """
    Causal Discovery using GES method.
    Link: https://pypi.org/project/ges/
    """

    def __init__(self, data, full_method_name, *args, **kwargs):
        super().__init__(data, full_method_name, *args, **kwargs)

        library_class = import_module(full_method_name)
        self._method = library_class

    def learn_graph(self, labels=None):
        """
        Discover causal graph and return the graph in DOT format.

        """
        self._adjacency_matrix, self.score = self._method.fit_bic(self._data.to_numpy())
        self._adjacency_matrix = np.asarray(self._adjacency_matrix)

        # If labels provided
        if labels is not None:
            self._labels = labels

        self._graph_dot = adjacency_matrix_to_graph(self._adjacency_matrix, self._labels)

        # Return in valid DOT format
        self._graph_dot = str_to_dot(self._graph_dot.source)
        return self._graph_dot



import copy
import logging

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import statsmodels.api as sm

from dowhy.causal_estimator import CausalEstimate, CausalEstimator
from dowhy.causal_estimators.econml import Econml
from dowhy.causal_estimators.generalized_linear_model_estimator import GeneralizedLinearModelEstimator
from dowhy.causal_estimators.linear_regression_estimator import LinearRegressionEstimator
from dowhy.causal_identifier import IdentifiedEstimand


class EValueSensitivityAnalyzer:
    """
    This class computes Ding & VanderWeele's E-value for unmeasured confounding. The E-value is the minimum
    strength of association on the risk ratio scale that an unmeasured confounder would need to have with
    both the treatment and the outcome, conditional on the measured covariates, to fully explain away a specific
    treatment-outcome association.

    It benchmarks the E-value against measured confounders using McGowan and Greevy Jr.'s Observed Covariate E-value.
    This approach drops measured confounders and re-fits the estimator, measuring how much the limiting bound of the
    confidence interval changes on the E-value scale. This benchmarks hypothetical unmeasured confounding against each
    of the measured confounders.

    See: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4820664/,
    https://dash.harvard.edu/bitstream/handle/1/36874927/EValue_FinalSubmission.pdf, and
    https://arxiv.org/pdf/2011.07030.pdf.
    The implementation is based on the R packages https://github.com/cran/EValue and https://github.com/LucyMcGowan/tipr.

    :param estimate: CausalEstimate
    :param estimand: IdentifiedEstimand
    :param data: pd.DataFrame
    :param outcome_name: Outcome variable name
    :param no_effect_baseline: A number to which to shift the observed estimate to. Defaults to 1 for ratio measures (RR, OR, HR)
    and 0 for additive measures (OLS, MD). (Default = None)
    """

    def __init__(
        self,
        estimate: CausalEstimate,
        estimand: IdentifiedEstimand,
        data: pd.DataFrame,
        treatment_name: str,
        outcome_name: str,
        no_effect_baseline=None,
    ):

        self.estimate = estimate
        self.estimand = estimand
        self.data = data
        self.treatment_name = treatment_name
        self.outcome_name = outcome_name
        self.no_effect_baseline = no_effect_baseline
        if self.no_effect_baseline is None:
            if isinstance(self.estimate.estimator, LinearRegressionEstimator):
                self.no_effect_baseline = 0
            elif isinstance(self.estimate.estimator, GeneralizedLinearModelEstimator):
                self.no_effect_baseline = 1
        self.logger = logging.getLogger(__name__)
        self.stats = None
        self.benchmarking_results = None
        self.sd_outcome = np.std(self.data[outcome_name])

    def check_sensitivity(self, plot=True):
        """
        Computes E-value for point estimate and confidence limits. Benchmarks E-values against
        measured confounders using Observed Covariate E-values. Plots E-values and Observed
        Covariate E-values.

        :param plot: plots E-value for point estimate and confidence limit. (Default = True)
        """

        if isinstance(self.estimate.estimator, LinearRegressionEstimator):
            coef_est = self.estimate.value
            coef_se = self.estimate.get_standard_error()[0]
        elif isinstance(self.estimate.estimator, GeneralizedLinearModelEstimator):
            coef_est = self.estimate.estimator.model.params[1:2].to_numpy()[0]
            coef_se = self.estimate.estimator.model.bse[1:2].to_numpy()[0]
        self.stats = self.get_evalue(coef_est, coef_se)

        eval_lo = self.stats["evalue_lower_ci"]
        eval_hi = self.stats["evalue_upper_ci"]
        if (eval_lo is None and eval_hi == 1) or (eval_hi is None and eval_lo == 1):
            self.logger.info(
                "Not benchmarking with Observed Covariate E-values. Confidence interval is already tipped."
            )
        else:
            self.benchmark()

        if plot:
            self.plot()

    def get_evalue(self, coef_est, coef_se):
        """
        Computes E-value for point estimate and confidence limits. The estimate and
        confidence limits are converted to the risk ratio scale before the E-value is calculated.

        :param coef_est: coefficient estimate
        :param coef_se: coefficient standard error
        """

        if isinstance(self.estimate.estimator, LinearRegressionEstimator):
            return self._evalue_OLS(coef_est, coef_se, self.sd_outcome, self.no_effect_baseline)
        elif isinstance(self.estimate.estimator, GeneralizedLinearModelEstimator):
            est = np.exp(coef_est)
            lo = np.exp(coef_est - 1.96 * coef_se)
            hi = np.exp(coef_est + 1.96 * coef_se)

            family = self.estimate.estimator.family
            link = self.estimate.estimator.family.link
            if isinstance(family, sm.families.Binomial) and isinstance(link, sm.families.links.Logit):
                rare = self.data[self.outcome_name].mean() < 0.15
                return self._evalue_OR(est, lo, hi, rare, self.no_effect_baseline)
            elif isinstance(
                family,
                (sm.families.Poisson, sm.families.NegativeBinomial, sm.families.Gamma),
            ) and isinstance(link, sm.families.links.Log):
                return self._evalue_RR(est, lo, hi, self.no_effect_baseline)
            else:
                raise NotImplementedError(
                    "Currently, only four GLM families are supported: "
                    "1) sm.families.Binomial(link=sm.families.links.logit), "
                    "2) sm.families.Poisson(link=sm.families.links.log), "
                    "3) sm.families.NegativeBinomial(link=sm.families.links.log), "
                    "4) sm.families.Gamma(link=sm.families.links.log), "
                )

    def plot(
        self,
        num_points_per_contour=200,
        plot_size=(6.4, 4.8),
        contour_colors=["blue", "red"],
        benchmarking_color="green",
        xy_limit=None,
    ):
        """
        Plots contours showing the combinations of treatment-confounder and confounder-outcome
        risk ratios that would tip the point estimate and confidence limit. The X-axis shows
        the treatment-confounder risk ratio and the Y-axis shows the confounder-outcome risk ratio.

        :param num_points_per_contour: number of points to calculate and plot for each contour (Default = 200)
        :param plot_size: size of the plot (Default = (6.4,4.8))
        :param contour_colors: colors for point estimate and confidence limit contour (Default = ["blue", "red"])
        :param benchmarking_color: color for observed covariate E-values. (Default = "green")
        :param xy_limit: plot's maximum x and y value. Default is 2 x E-value. (Default = None)
        """

        if self.stats is None:
            raise ValueError("Must call ``check_sensitivity'' before calling ``plot''")

        fig, ax = plt.subplots(1, 1, figsize=plot_size)

        eval_est = self.stats["evalue_estimate"]
        rr = self.stats["converted_estimate"]
        if rr < 1:
            rr = 1 / rr

        if xy_limit is None:
            xy_limit = eval_est * 2

        self._plot_contour(ax, rr, eval_est, num_points_per_contour, contour_colors[0], xy_limit)

        eval_lo = self.stats["evalue_lower_ci"]
        eval_hi = self.stats["evalue_upper_ci"]
        if (eval_lo is None and eval_hi == 1) or (eval_hi is None and eval_lo == 1):
            self.logger.info("Plotting contour for point estimate only. Confidence interval is already tipped.")
        else:
            if eval_lo is None:
                rr_ci = self.stats["converted_upper_ci"]
                eval_ci = eval_hi
            else:
                rr_ci = self.stats["converted_lower_ci"]
                eval_ci = eval_lo
            if rr_ci < 1:
                rr_ci = 1 / rr_ci

            self._plot_contour(
                ax,
                rr_ci,
                eval_ci,
                num_points_per_contour,
                contour_colors[1],
                xy_limit,
                point_est=False,
            )

            ax.scatter(
                self.benchmarking_results["observed_covariate_e_value"],
                self.benchmarking_results["observed_covariate_e_value"],
                label="Observed Covariate E-values",
                color=benchmarking_color,
            )
            example_var = self.benchmarking_results.iloc[0]
            obs_evalue = example_var["observed_covariate_e_value"]
            ax.text(obs_evalue, obs_evalue, example_var.name)

        ax.set(xlabel="$RR_{treatment-confounder}$", ylabel="$RR_{confounder-outcome}$")
        plt.ylim(1, xy_limit)
        plt.xlim(1, xy_limit)
        plt.legend()
        plt.show()

    def _plot_contour(self, ax, rr, evalue, n_pts, color, xy_limit, point_est=True):
        """
        Plots a single contour line

        :param ax: matplotlib axis
        :param rr: observed point estimate/confidence limit on the risk ratio scale
        :param evalue: E-value corresponding to observed point estimate/confidence limit
        :param n_pts: number of points to calculate and plot for each contour
        :param color: color for contour
        :param xy_limit: plot's maximum x and y value
        :param point_est: whether this is the point estimate (Default = True)
        """

        step = (xy_limit - rr) / n_pts
        x_est = np.linspace(rr + step, xy_limit, num=n_pts)
        y_est = rr * (rr - 1) / (x_est - rr) + rr

        est_string = "point estimate" if point_est else "confidence interval"
        ax.scatter(
            evalue,
            evalue,
            label=f"E-value for {est_string}: {evalue.round(2)}",
            color=color,
        )
        ax.fill_between(
            x_est,
            y_est,
            xy_limit,
            color=color,
            alpha=0.2,
            label=f"Tips {est_string}",
        )
        ax.plot(x_est, y_est, color=color)

    def benchmark(self):
        """
        Benchmarks E-values against the measured confounders using McGowan and Greevy Jr.'s Observed
        Covariate E-value. This approach drops measured confounders and re-fits the estimator, measuring
        how much the limiting bound of the confidence interval changes on the E-value scale. This benchmarks
        hypothetical unmeasured confounding against each of the measured confounders.

        See: https://arxiv.org/pdf/2011.07030.pdf and https://github.com/LucyMcGowan/tipr
        """

        new_ests = []
        new_lo = []
        new_hi = []
        observed_covariate_e_values = []
        backdoor_vars = self.estimand.get_backdoor_variables()
        for drop_var in backdoor_vars:

            # new estimator
            new_backdoor_vars = [var for var in backdoor_vars if var != drop_var]
            new_estimand = copy.deepcopy(self.estimand)
            new_estimand.set_backdoor_variables(new_backdoor_vars)
            new_estimator = self.estimate.estimator.get_new_estimator_object(new_estimand)
            new_estimator.fit(
                self.data,
                new_estimand.treatment_variable,
                new_estimand.outcome_variable,
                self.estimate.estimator._effect_modifier_names,
                **new_estimator._econml_fit_params if isinstance(new_estimator, Econml) else {},
            )

            # new effect estimate
            new_effect = new_estimator.estimate_effect(
                control_value=self.estimate.control_value,
                treatment_value=self.estimate.treatment_value,
                target_units=self.estimate.estimator._target_units,
            )
            if isinstance(self.estimate.estimator, LinearRegressionEstimator):
                coef_est = new_effect.value
                coef_se = new_effect.get_standard_error()[0]
            elif isinstance(self.estimate.estimator, GeneralizedLinearModelEstimator):
                coef_est = new_effect.estimator.model.params[1:2].to_numpy()[0]
                coef_se = new_effect.estimator.model.bse[1:2].to_numpy()[0]
            new_stats = self.get_evalue(coef_est, coef_se)

            # observed covariate E-value
            if self.stats["evalue_lower_ci"] is None:
                ci = self.stats["converted_upper_ci"]
                new_ci = new_stats["converted_upper_ci"]
            else:
                ci = self.stats["converted_lower_ci"]
                new_ci = new_stats["converted_lower_ci"]
            covariate_e_value = self._observed_covariate_e_value(ci, new_ci)

            new_ests.append(new_stats["converted_estimate"])
            new_lo.append(new_stats["converted_lower_ci"])
            new_hi.append(new_stats["converted_upper_ci"])
            observed_covariate_e_values.append(covariate_e_value)

        self.benchmarking_results = pd.DataFrame(
            {
                "dropped_covariate": backdoor_vars,
                "converted_est": new_ests,
                "converted_lower_ci": new_lo,
                "converted_upper_ci": new_hi,
                "observed_covariate_e_value": observed_covariate_e_values,
            }
        ).sort_values(by="observed_covariate_e_value", ascending=False)
        self.benchmarking_results = self.benchmarking_results.set_index("dropped_covariate")

    def _observed_covariate_e_value(self, ci, new_ci):
        """
        Computes Observed Covariate E-value given effect estimate from new model without
        a specific measured confounder.

        Based on: https://github.com/LucyMcGowan/tipr/blob/master/R/observed_covariate_e_value.R

        :param ci: limiting confidence bound from original model on risk ratio scale
        :param new_ci: limiting confidence bound from new model on risk ratio scale
        """

        if ci < 1:
            ci = 1 / ci
            new_ci = 1 / new_ci
        if ci < new_ci:
            ratio = new_ci / ci
        else:
            ratio = ci / new_ci

        return ratio + np.sqrt(ratio * (ratio - 1))

    def _evalue_OLS(self, est, se, sd, no_effect_baseline=0):
        """
        Computes E-value from OLS coefficient.

        :param est: coefficient from OLS
        :param se: standard error of point estimate
        :param sd: residual standard deviation
        :param no_effect_baseline: no_effect_baseline standardized difference to which to shift the observed estimate. (Default = 0)
        """
        if se < 0:
            raise ValueError("Standard error cannot be negative")

        delta = abs(self.estimate.estimator._treatment_value - self.estimate.estimator._control_value)
        est = self._to_smd(est, sd, delta=delta)
        se = self._to_smd(
            se, sd, delta=1
        )  # already multiplied by delta in LinearRegressionEstimator._estimate_std_error()
        return self._evalue_MD(est=est, se=se, no_effect_baseline=no_effect_baseline)

    def _evalue_MD(self, est, se=None, no_effect_baseline=0):
        """
        Computes E-value for standardized difference and its confidence limits.

        :param est: point estimate as standardized difference (i.e., Cohen's d)
        :param se: standard error of the point estimate
        :param no_effect_baseline: no_effect_baseline standardized difference to which to shift the observed estimate. (Default = 0)
        """
        if se < 0:
            raise ValueError("Standard error cannot be negative")

        if se is None:
            lo = None
            hi = None
        else:
            # see Table 2 and p.37 in https://dash.harvard.edu/bitstream/handle/1/36874927/EValue_FinalSubmission.pdf
            lo = np.exp(0.91 * est - 1.78 * se)
            hi = np.exp(0.91 * est + 1.78 * se)

        est = self._md_to_rr(est)
        no_effect_baseline = self._md_to_rr(no_effect_baseline)

        return self._evalue_RR(est=est, lo=lo, hi=hi, no_effect_baseline=no_effect_baseline)

    def _evalue_OR(self, est, lo=None, hi=None, rare=None, no_effect_baseline=1):
        """
        Computes E-value for an odds ratio and its confidence limits.

        :param est: point estimate
        :param lo: lower limit of confidence interval
        :param hi: upper limit of confidence interval
        :param rare: if outcome is rare (<15%)
        :param no_effect_baseline: the no_effect_baseline OR to which to shift the observed estimate. (Default = 1)
        """
        if est < 0:
            raise ValueError("Odds Ratio cannot be negative")

        est = self._or_to_rr(est, rare)
        if lo is not None:
            lo = self._or_to_rr(lo, rare)
        if hi is not None:
            hi = self._or_to_rr(hi, rare)
        no_effect_baseline = self._or_to_rr(no_effect_baseline, rare)

        return self._evalue_RR(est=est, lo=lo, hi=hi, no_effect_baseline=no_effect_baseline)

    def _evalue_RR(self, est, lo=None, hi=None, no_effect_baseline=1):
        """
        Computes E-value for a risk ratio or rate ratio and its confidence limits.

        :param est: point estimate
        :param lo: lower limit of confidence interval
        :param hi: upper limit of confidence interval
        :param no_effect_baseline: the no_effect_baseline RR to which to shift the observed estimate. (Default = 1)
        """
        if est < 0:
            raise ValueError("Risk/Rate Ratio cannot be negative")
        if no_effect_baseline < 0:
            raise ValueError("no_effect_baseline value is impossible")
        if no_effect_baseline != 1:
            self.logger.info(
                'You are calculating a "non-null" E-value, i.e., an E-value for the minimum amount of unmeasured '
                "confounding needed to move the estimate and confidence interval to your specified no_effect_baseline value "
                "rather than to the null value."
            )
        if lo is not None and hi is not None:
            if lo > hi:
                raise ValueError("Lower confidence limit should be less than upper confidence limit")
        if lo is not None and est < lo:
            raise ValueError("Point estimate should be inside confidence interval")
        if hi is not None and est > hi:
            raise ValueError("Point estimate should be inside confidence interval")

        e_est = self._threshold(est, no_effect_baseline=no_effect_baseline)
        e_lo = self._threshold(lo, no_effect_baseline=no_effect_baseline)
        e_hi = self._threshold(hi, no_effect_baseline=no_effect_baseline)

        # if CI crosses null, set its E-value to 1
        null_CI = False
        if est > no_effect_baseline and lo is not None:
            null_CI = lo < no_effect_baseline
        if est < no_effect_baseline and hi is not None:
            null_CI = hi > no_effect_baseline
        if null_CI:
            e_lo = np.float64(1)
            e_hi = np.float64(1)

        # only report E-value for CI limit closer to null
        if lo is not None or hi is not None:
            if est > no_effect_baseline:
                e_hi = None
            else:
                e_lo = None

        return {
            "converted_estimate": est,
            "converted_lower_ci": lo,
            "converted_upper_ci": hi,
            "evalue_estimate": e_est,
            "evalue_lower_ci": e_lo,
            "evalue_upper_ci": e_hi,
        }

    def _threshold(self, x, no_effect_baseline=1):
        """
        Computes E-value for single value of risk ratio.

        :param x: risk ratio
        :param no_effect_baseline: the no_effect_baseline RR to which to shift the observed estimate
        """
        if x is None:
            return None

        if x <= 1:
            x = 1 / x
            no_effect_baseline = 1 / no_effect_baseline

        if no_effect_baseline <= x:
            return (x + np.sqrt(x * (x - no_effect_baseline))) / no_effect_baseline
        else:
            ratio = no_effect_baseline / x
            return ratio + np.sqrt(ratio * (ratio - 1))

    def _to_smd(self, est, sd, delta=1):
        """
        Converts estimate to standardized mean difference.

        :param est: estimate
        :param sd: residual standard deviation or standard deviation of outcome
        :param delta: contrast of interest in the treatment/exposure
        """
        return est * delta / sd

    def _md_to_rr(self, est):
        """
        Converts standardized mean difference to approximate risk ratio.

        :param est: estimate
        """
        # see Table 2 and p.37 in https://dash.harvard.edu/bitstream/handle/1/36874927/EValue_FinalSubmission.pdf
        return np.exp(0.91 * est)

    def _hr_to_rr(self, est, rare):
        """
        Converts hazard ratio to approximate risk ratio.

        :param est: estimate
        :param rare: if outcome is rare (<15%)
        """
        if rare is None:
            raise ValueError('Must specify whether the rare outcome assumption can be made. Use argument "rare" =')
        if rare:
            return est
        else:
            return (1 - 0.5 ** np.sqrt(est)) / ((1 - 0.5 ** np.sqrt(1 / est)))

    def _or_to_rr(self, est, rare):
        """
        Converts odds ratio to approximate risk ratio.

        :param est: estimate
        :param rare: if outcome is rare (<15%)
        """
        if rare:
            return est
        else:
            return np.sqrt(est)

    def __str__(self):
        s = "Sensitivity Analysis to Unobserved Confounding using the E-value\n\n"
        s += f"Unadjusted Estimates of Treatment: {self.treatment_name}\n"
        s += f"Estimate (converted to risk ratio scale): {self.stats['converted_estimate']}\n"
        s += f"Lower 95% CI (converted to risk ratio scale): {self.stats['converted_lower_ci']}\n"
        s += f"Upper 95% CI (converted to risk ratio scale): {self.stats['converted_upper_ci']}\n\n"
        s += "Sensitivity Statistics: \n"
        s += f"E-value for point estimate: {self.stats['evalue_estimate']}\n"
        s += f"E-value for lower 95% CI: {self.stats['evalue_lower_ci']}\n"
        s += f"E-value for upper 95% CI: {self.stats['evalue_upper_ci']}\n"
        largest_obs_evalue = self.benchmarking_results["observed_covariate_e_value"].iloc[0]
        largest_obs_cov = self.benchmarking_results.index[0]
        s += f"Largest Observed Covariate E-value: {largest_obs_evalue} ({largest_obs_cov})\n\n"
        s += "Interpretation of results:\n"
        if self.stats["evalue_lower_ci"] is None:
            ci = self.stats["evalue_upper_ci"]
            direction = "decrease"
        else:
            ci = self.stats["evalue_lower_ci"]
            direction = "increase"
        s += (
            f"Unmeasured confounder(s) would have to be associated with a {self.stats['evalue_estimate'].round(2)}-fold {direction} "
            f"in the risk of {self.outcome_name}, and must be {self.stats['evalue_estimate'].round(2)} times more prevalent in "
            f"{self.treatment_name}, to explain away the observed point estimate.\n"
        )
        s += (
            f"Unmeasured confounder(s) would have to be associated with a {ci.round(2)}-fold {direction} "
            f"in the risk of {self.outcome_name}, and must be {ci.round(2)} times more prevalent in "
            f"{self.treatment_name}, to explain away the observed confidence interval."
        )
        return s



from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd
import sympy as sp
import sympy.stats as spstats
from statsmodels.sandbox.regression.gmm import IV2SLS

from dowhy.causal_estimator import CausalEstimate, CausalEstimator, RealizedEstimand
from dowhy.causal_identifier import IdentifiedEstimand
from dowhy.utils.api import parse_state


class InstrumentalVariableEstimator(CausalEstimator):
    """Compute effect of treatment using the instrumental variables method.

    This is also a superclass that can be inherited by other specific methods.

    Supports additional parameters as listed below.

    """

    def __init__(
        self,
        identified_estimand: IdentifiedEstimand,
        iv_instrument_name: Optional[Union[List, Dict, str]] = None,
        test_significance: bool = False,
        evaluate_effect_strength: bool = False,
        confidence_intervals: bool = False,
        num_null_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_STAT_TEST,
        num_simulations: int = CausalEstimator.DEFAULT_NUMBER_OF_SIMULATIONS_CI,
        sample_size_fraction: int = CausalEstimator.DEFAULT_SAMPLE_SIZE_FRACTION,
        confidence_level: float = CausalEstimator.DEFAULT_CONFIDENCE_LEVEL,
        need_conditional_estimates: Union[bool, str] = "auto",
        num_quantiles_to_discretize_cont_cols: int = CausalEstimator.NUM_QUANTILES_TO_DISCRETIZE_CONT_COLS,
        **kwargs,
    ):
        """
        :param identified_estimand: probability expression
            representing the target identified estimand to estimate.
        :param iv_instrument_name: Name of the specific instrumental variable
            to be used. Needs to be one of the IVs identified in the
            identification step. Default is to use all the IV variables
            from the identification step.
        :param test_significance: Binary flag or a string indicating whether to test significance and by which method. All estimators support test_significance="bootstrap" that estimates a p-value for the obtained estimate using the bootstrap method. Individual estimators can override this to support custom testing methods. The bootstrap method supports an optional parameter, num_null_simulations. If False, no testing is done. If True, significance of the estimate is tested using the custom method if available, otherwise by bootstrap.
        :param evaluate_effect_strength: (Experimental) whether to evaluate the strength of effect
        :param confidence_intervals: Binary flag or a string indicating whether the confidence intervals should be computed and which method should be used. All methods support estimation of confidence intervals using the bootstrap method by using the parameter confidence_intervals="bootstrap". The bootstrap method takes in two arguments (num_simulations and sample_size_fraction) that can be optionally specified in the params dictionary. Estimators may also override this to implement their own confidence interval method. If this parameter is False, no confidence intervals are computed. If True, confidence intervals are computed by the estimator's specific method if available, otherwise through bootstrap
        :param num_null_simulations: The number of simulations for testing the
            statistical significance of the estimator
        :param num_simulations: The number of simulations for finding the
            confidence interval (and/or standard error) for a estimate
        :param sample_size_fraction: The size of the sample for the bootstrap
            estimator
        :param confidence_level: The confidence level of the confidence
            interval estimate
        :param need_conditional_estimates: Boolean flag indicating whether
            conditional estimates should be computed. Defaults to True if
            there are effect modifiers in the graph
        :param num_quantiles_to_discretize_cont_cols: The number of quantiles
            into which a numeric effect modifier is split, to enable
            estimation of conditional treatment effect over it.
        :param kwargs: (optional) Additional estimator-specific parameters
        """
        super().__init__(
            identified_estimand=identified_estimand,
            test_significance=test_significance,
            evaluate_effect_strength=evaluate_effect_strength,
            confidence_intervals=confidence_intervals,
            num_null_simulations=num_null_simulations,
            num_simulations=num_simulations,
            sample_size_fraction=sample_size_fraction,
            confidence_level=confidence_level,
            need_conditional_estimates=need_conditional_estimates,
            num_quantiles_to_discretize_cont_cols=num_quantiles_to_discretize_cont_cols,
            iv_instrument_name=iv_instrument_name,
            **kwargs,
        )
        self.iv_instrument_name = iv_instrument_name
        self.logger.info("INFO: Using Instrumental Variable Estimator")

    def fit(
        self,
        data: pd.DataFrame,
        treatment_name: str,
        outcome_name: str,
        effect_modifier_names: Optional[List[str]] = None,
    ):
        """
        Fits the estimator with data for effect estimation
        :param data: data frame containing the data
        :param treatment: name of the treatment variable
        :param outcome: name of the outcome variable
        :param effect_modifiers: Variables on which to compute separate
                    effects, or return a heterogeneous effect function. Not all
                    methods support this currently.
        """
        self._set_data(data, treatment_name, outcome_name)
        self._set_effect_modifiers(effect_modifier_names)

        self.estimating_instrument_names = self._target_estimand.instrumental_variables
        if self.iv_instrument_name is not None:
            self.estimating_instrument_names = parse_state(self.iv_instrument_name)
        self.logger.debug("Instrumental Variables used:" + ",".join(self.estimating_instrument_names))
        if not self.estimating_instrument_names:
            raise ValueError("No valid instruments found. IV Method not applicable")
        if len(self.estimating_instrument_names) < len(self._treatment_name):
            # TODO move this to the identification step
            raise ValueError(
                "Number of instruments fewer than number of treatments. 2SLS requires at least as many instruments as treatments."
            )
        self._estimating_instruments = self._data[self.estimating_instrument_names]

        self.symbolic_estimator = self.construct_symbolic_estimator(self._target_estimand)
        self.logger.info(self.symbolic_estimator)

        return self

    def estimate_effect(
        self, data: pd.DataFrame = None, treatment_value: Any = 1, control_value: Any = 0, target_units=None, **_
    ):
        """
        data: dataframe containing the data on which treatment effect is to be estimated.
        treatment_value: value of the treatment variable for which the effect is to be estimated.
        control_value: value of the treatment variable that denotes its absence (usually 0)
        target_units: The units for which the treatment effect should be estimated.
                     It can be a DataFrame that contains values of the effect_modifiers and effect will be estimated only for this new data.
                     It can also be a lambda function that can be used as an index for the data (pandas DataFrame) to select the required rows.
        """
        if data is None:
            data = self._data
        self._target_units = target_units
        self._treatment_value = treatment_value
        self._control_value = control_value
        if len(self.estimating_instrument_names) == 1 and len(self._treatment_name) == 1:
            instrument = self._estimating_instruments.iloc[:, 0]
            self.logger.debug("Instrument Variable values: {0}".format(instrument))
            num_unique_values = len(np.unique(instrument))
            instrument_is_binary = num_unique_values <= 2
            if instrument_is_binary:
                # Obtain estimate by Wald Estimator
                y1_z = np.mean(self._outcome[instrument == 1])
                y0_z = np.mean(self._outcome[instrument == 0])
                x1_z = np.mean(self._treatment[self._treatment_name[0]][instrument == 1])
                x0_z = np.mean(self._treatment[self._treatment_name[0]][instrument == 0])
                num = y1_z - y0_z
                deno = x1_z - x0_z
                iv_est = num / deno
            else:
                # Obtain estimate by 2SLS estimator: Cov(y,z) / Cov(x,z)
                num_yz = np.cov(self._outcome, instrument)[0, 1]
                deno_xz = np.cov(self._treatment[self._treatment_name[0]], instrument)[0, 1]
                iv_est = num_yz / deno_xz
        else:
            # More than 1 instrument. Use 2sls.
            est_treatment = self._treatment.astype(np.float32)
            est_outcome = self._outcome.astype(np.float32)
            ivmodel = IV2SLS(est_outcome, est_treatment, self._estimating_instruments)
            reg_results = ivmodel.fit()
            self.logger.debug(reg_results.summary())
            iv_est = sum(
                reg_results.params
            )  # the effect is the same for any treatment value (assume treatment goes from 0 to 1)
        estimate = CausalEstimate(
            estimate=iv_est,
            control_value=control_value,
            treatment_value=treatment_value,
            target_estimand=self._target_estimand,
            realized_estimand_expr=self.symbolic_estimator,
        )

        estimate.add_estimator(self)
        return estimate

    def construct_symbolic_estimator(self, estimand):
        sym_outcome = spstats.Normal(",".join(estimand.outcome_variable), 0, 1)
        sym_treatment = spstats.Normal(",".join(estimand.treatment_variable), 0, 1)
        sym_instrument = sp.Symbol(",".join(self.estimating_instrument_names))
        sym_outcome_derivative = sp.Derivative(sym_outcome, sym_instrument)
        sym_treatment_derivative = sp.Derivative(sym_treatment, sym_instrument)
        sym_effect = spstats.Expectation(sym_outcome_derivative) / sp.stats.Expectation(sym_treatment_derivative)
        estimator_assumptions = {
            "treatment_effect_homogeneity": (
                "Each unit's treatment {0} is ".format(self._treatment_name)
                + "affected in the same way by common causes of "
                "{0} and {1}".format(self._treatment_name, self._outcome_name)
            ),
            "outcome_effect_homogeneity": (
                "Each unit's outcome {0} is ".format(self._outcome_name)
                + "affected in the same way by common causes of "
                "{0} and {1}".format(self._treatment_name, self._outcome_name)
            ),
        }
        sym_assumptions = {**estimand.estimands["iv"]["assumptions"], **estimator_assumptions}

        symbolic_estimand = RealizedEstimand(estimand, estimator_name="Wald Estimator")
        symbolic_estimand.update_assumptions(sym_assumptions)
        symbolic_estimand.update_estimand_expression(sym_effect)
        return symbolic_estimand
